@@ -1,7 +1,6 @@
 //=---- Internals.h - Interface between the backend components ----*- C++ -*-=//
 //
-// Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011  Chris Lattner,
-// Duncan Sands et al.
+// Copyright (C) 2005 to 2012  Chris Lattner, Duncan Sands et al.
 //
 // This file is part of DragonEgg.
 //
@@ -38,11 +37,6 @@ struct basic_block_def;
 union gimple_statement_d;
 union tree_node;
 
-extern "C" {
-extern void debug_gimple_stmt(union gimple_statement_d *);
-extern void debug_tree(union tree_node *);
-}
-
 namespace llvm {
   class Module;
   class GlobalVariable;
@@ -58,10 +52,10 @@ namespace llvm {
   class Type;
   class TargetMachine;
   class TargetData;
-  class DebugInfo;
   template<typename> class AssertingVH;
   template<typename> class TrackingVH;
 }
+class DebugInfo;
 using namespace llvm;
 
 typedef IRBuilder<true, TargetFolder> LLVMBuilder;
@@ -74,7 +68,7 @@ extern llvm::Module *TheModule;
 
 /// TheDebugInfo - This object is responsible for gather all debug information.
 /// If it's value is NULL then no debug information should be gathered.
-extern llvm::DebugInfo *TheDebugInfo;
+extern DebugInfo *TheDebugInfo;
 
 /// TheTarget - The current target being compiled for.
 ///
@@ -106,23 +100,6 @@ extern SmallSetVector<Constant *,32> AttributeUsedGlobals;
 
 extern Constant* ConvertMetadataStringToGV(const char *str);
 
-/// DieAbjectly - An unrecoverable fatal error occurred - throw in the towel,
-/// give up the ghost, quit miserably.
-inline void LLVM_ATTRIBUTE_NORETURN DieAbjectly(const char *Message) {
-  llvm_unreachable(Message);
-  (void)Message; // Avoid unused variable warning when assertions are disabled.
-}
-inline void LLVM_ATTRIBUTE_NORETURN DieAbjectly(const char *Message,
-                                                union gimple_statement_d *stmt){
-  if (stmt) debug_gimple_stmt(stmt);
-  DieAbjectly(Message);
-}
-inline void LLVM_ATTRIBUTE_NORETURN DieAbjectly(const char *Message,
-                                                union tree_node *exp) {
-  if (exp) debug_tree(exp);
-  DieAbjectly(Message);
-}
-
 /// AddAnnotateAttrsToGlobal - Adds decls that have a
 /// annotate attribute to a vector to be emitted later.
 extern void AddAnnotateAttrsToGlobal(GlobalValue *GV, tree_node *decl);
@@ -159,7 +136,7 @@ void changeLLVMConstant(Constant *Old, Constant *New);
 void register_ctor_dtor(Function *, int, bool);
 const char *extractRegisterName(tree_node *);
 void handleVisibility(tree_node *decl, GlobalValue *GV);
-Twine getLLVMAssemblerName(tree_node *);
+std::string getLLVMAssemblerName(tree_node *);
 
 /// Return true if and only if field no. N from struct type T is a padding
 /// element added to match llvm struct type size and gcc struct type size.
@@ -194,9 +171,7 @@ private:
 public:
   explicit MemRef() : Ptr(0), Volatile(false), LogAlign(0) {}
   explicit MemRef(Value *P, uint32_t A, bool V) : Ptr(P), Volatile(V) {
-    // Forbid alignment 0 along with non-power-of-2 alignment values.
-    assert(isPowerOf2_32(A) && "Alignment not a power of 2!");
-    LogAlign = Log2_32(A);
+    setAlignment(A);
   }
 
   uint32_t getAlignment() const {
@@ -204,7 +179,9 @@ public:
   }
 
   void setAlignment(uint32_t A) {
-    LogAlign = Log2_32(A);
+    // Forbid alignment 0 along with non-power-of-2 alignment values.
+    assert(isPowerOf2_32(A) && "Alignment not a power of 2!");
+    LogAlign = (unsigned char)Log2_32(A);
   }
 };
 
@@ -225,7 +202,8 @@ public:
   LValue(Value *P, uint32_t A, bool V = false) :
       MemRef(P, A, V), BitStart(255), BitSize(255) {}
   LValue(Value *P, uint32_t A, unsigned BSt, unsigned BSi, bool V = false) :
-      MemRef(P, A, V), BitStart(BSt), BitSize(BSi) {
+      MemRef(P, A, V), BitStart((unsigned char)BSt),
+      BitSize((unsigned char)BSi) {
     assert(BitStart == BSt && BitSize == BSi &&
            "Bit values larger than 256?");
   }
@@ -350,15 +328,13 @@ public:
   /// of the types involved. This is an inferred cast.
   Value *CastToAnyType (Value *V, bool VSigned, Type *Ty, bool TySigned);
 
-  /// CastToUIntType - Cast the specified value to the specified type assuming
-  /// that V's type and Ty are integral types. This arbitrates between BitCast,
-  /// Trunc and ZExt.
-  Value *CastToUIntType(Value *V, Type *Ty);
+  /// CastFromSameSizeInteger - Cast an integer (or vector of integer) value to
+  /// the given scalar (resp. vector of scalar) type of the same bitwidth.
+  Value *CastFromSameSizeInteger(Value *V, Type *Ty);
 
-  /// CastToSIntType - Cast the specified value to the specified type assuming
-  /// that V's type and Ty are integral types. This arbitrates between BitCast,
-  /// Trunc and SExt.
-  Value *CastToSIntType(Value *V, Type *Ty);
+  /// CastToSameSizeInteger - Cast the specified scalar (or vector of scalar)
+  /// value to an integer (resp. vector of integer) of the same bit width.
+  Value *CastToSameSizeInteger(Value *V);
 
   /// CastToFPType - Cast the specified value to the specified type assuming
   /// that V's type and Ty are floating point types. This arbitrates between
@@ -407,6 +383,10 @@ private: // Helper functions.
   /// cleans up and returns the result function.
   Function *FinishFunctionBody();
 
+  /// EmitVariablesInScope - Output a declaration for every variable in the
+  /// given scope.
+  void EmitVariablesInScope(tree_node *scope);
+
   /// PopulatePhiNodes - Populate generated phi nodes with their operands.
   void PopulatePhiNodes();
 
@@ -456,21 +436,17 @@ private: // Helper functions.
 
 private: // Helpers for exception handling.
 
-  /// getLandingPad - Return the landing pad for the given exception handling
-  /// region, creating it if necessary.
-  BasicBlock *getLandingPad(unsigned RegionNo);
-
   /// getExceptionPtr - Return the local holding the exception pointer for the
   /// given exception handling region, creating it if necessary.
-  AllocaInst *getExceptionPtr(unsigned RegionNo);
+  AllocaInst *getExceptionPtr(int RegionNo);
 
   /// getExceptionFilter - Return the local holding the filter value for the
   /// given exception handling region, creating it if necessary.
-  AllocaInst *getExceptionFilter(unsigned RegionNo);
+  AllocaInst *getExceptionFilter(int RegionNo);
 
   /// getFailureBlock - Return the basic block containing the failure code for
   /// the given exception handling region, creating it if necessary.
-  BasicBlock *getFailureBlock(unsigned RegionNo);
+  BasicBlock *getFailureBlock(int RegionNo);
 
 private:
   void EmitAutomaticVariableDecl(tree_node *decl);
@@ -579,10 +555,12 @@ private:
   Value *EmitReg_ROUND_DIV_EXPR(tree_node *op0, tree_node *op1);
   Value *EmitReg_TRUNC_DIV_EXPR(tree_node *op0, tree_node *op1, bool isExact);
   Value *EmitReg_TRUNC_MOD_EXPR(tree_node *op0, tree_node *op1);
+#if (GCC_MINOR < 7)
   Value *EmitReg_VEC_EXTRACT_EVEN_EXPR(tree_node *op0, tree_node *op1);
   Value *EmitReg_VEC_EXTRACT_ODD_EXPR(tree_node *op0, tree_node *op1);
   Value *EmitReg_VEC_INTERLEAVE_HIGH_EXPR(tree_node *op0, tree_node *op1);
   Value *EmitReg_VEC_INTERLEAVE_LOW_EXPR(tree_node *op0, tree_node *op1);
+#endif
   Value *EmitReg_VEC_PACK_TRUNC_EXPR(tree_node *type, tree_node *op0,
                                      tree_node *op1);
   Value *EmitReg_VEC_WIDEN_MULT_HI_EXPR(tree_node *type, tree_node *op0,
@@ -592,10 +570,18 @@ private:
   Value *EmitReg_WIDEN_MULT_EXPR(tree_node *type, tree_node *op0,
                                  tree_node *op1);
 
+  // Ternary expressions.
+  Value *EmitReg_CondExpr(tree_node *op0, tree_node *op1, tree_node *op2);
+#if (GCC_MINOR > 6)
+  Value *EmitReg_VEC_PERM_EXPR(tree_node *op0, tree_node *op1, tree_node *op2);
+#endif
+
   Value *EmitLoadOfLValue(tree_node *exp);
   Value *EmitOBJ_TYPE_REF(tree_node *exp);
   Value *EmitADDR_EXPR(tree_node *exp);
+#if (GCC_MINOR < 7)
   Value *EmitCondExpr(tree_node *exp);
+#endif
   Value *EmitCallOf(Value *Callee, gimple_statement_d *stmt,
                     const MemRef *DestLoc, const AttrListPtr &PAL);
   CallInst *EmitSimpleCall(StringRef CalleeName, tree_node *ret_type,
@@ -622,6 +608,8 @@ private:
                                        tree_node *fndecl, const MemRef *DestLoc,
                                        Value *&Result);
   bool EmitBuiltinUnaryOp(Value *InVal, Value *&Result, Intrinsic::ID Id);
+  Value *EmitBuiltinBitCountIntrinsic(gimple_statement_d *stmt,
+                                      Intrinsic::ID Id);
   Value *EmitBuiltinSQRT(gimple_statement_d *stmt);
   Value *EmitBuiltinPOWI(gimple_statement_d *stmt);
   Value *EmitBuiltinPOW(gimple_statement_d *stmt);
@@ -724,38 +712,11 @@ private:
   /// constant of the corresponding register type.
   Constant *EmitVectorRegisterConstant(tree_node *reg);
 
-  /// Mem2Reg - Convert a value of in-memory type (that given by ConvertType)
-  /// to in-register type (that given by getRegType).  TODO: Eliminate these
-  /// methods: "memory" values should never be held in registers.  Currently
-  /// this is mainly used for marshalling function parameters and return values,
-  /// but that should be completely independent of the reg vs mem value logic.
-  Value *Mem2Reg(Value *V, tree_node *type, LLVMBuilder &Builder);
-
-  /// Reg2Mem - Convert a value of in-register type (that given by getRegType)
-  /// to in-memory type (that given by ConvertType).  TODO: Eliminate this
-  /// method: "memory" values should never be held in registers.  Currently
-  /// this is mainly used for marshalling function parameters and return values,
-  /// but that should be completely independent of the reg vs mem value logic.
-  Value *Reg2Mem(Value *V, tree_node *type, LLVMBuilder &Builder);
-
   /// EmitMemory - Convert the specified gimple register or local constant of
   /// register type to an LLVM value with in-memory type (given by ConvertType).
   /// TODO: Eliminate this method, see Mem2Reg and Reg2Mem above.
   Value *EmitMemory(tree_node *reg);
 
-  /// LoadRegisterFromMemory - Loads a value of the given scalar GCC type from
-  /// the memory location pointed to by Loc.  Takes care of adjusting for any
-  /// differences between in-memory and in-register types (the returned value
-  /// is of in-register type, as returned by getRegType).
-  Value *LoadRegisterFromMemory(MemRef Loc, tree_node *type,
-                                LLVMBuilder &Builder);
-
-  /// StoreRegisterToMemory - Stores the given value to the memory pointed to by
-  /// Loc.  Takes care of adjusting for any differences between the value's type
-  /// (which is the in-register type given by getRegType) and the in-memory type.
-  void StoreRegisterToMemory(Value *V, MemRef Loc, tree_node *type,
-                             LLVMBuilder &Builder);
-
   /// VectorHighElements - Return a vector of half the length, consisting of the
   /// elements of the given vector with indices in the top half.
   Value *VectorHighElements(Value *Vec);
