@@ -1,7 +1,6 @@
 //==----- Target.h - Target hooks for GCC to LLVM conversion -----*- C++ -*-==//
 //
-// Copyright (C) 2007, 2008, 2009, 2010, 2011  Anton Korobeynikov, Duncan Sands
-// et al.
+// Copyright (C) 2007 to 2012  Anton Korobeynikov, Duncan Sands et al.
 //
 // This file is part of DragonEgg.
 //
@@ -47,7 +46,14 @@
    `inreg` parameter attribute */
 #define LLVM_TARGET_ENABLE_REGPARM
 
-extern "C" int ix86_regparm;
+#include "auto-host.h"
+#ifndef ENABLE_BUILD_WITH_CXX
+extern "C" {
+#endif
+extern int ix86_regparm;
+#ifndef ENABLE_BUILD_WITH_CXX
+} // extern "C"
+#endif
 
 #define LLVM_TARGET_INIT_REGPARM(local_regparm, local_fp_regparm, type) \
   {                                                             \
@@ -71,7 +77,7 @@ extern "C" int ix86_regparm;
                                       local_fp_regparm)         \
   {                                                             \
     if (!TARGET_64BIT) {                                        \
-      if (TREE_CODE(Type) == REAL_TYPE &&                       \
+      if (isa<REAL_TYPE>(Type) &&                               \
           (TYPE_PRECISION(Type)==32 ||                          \
            TYPE_PRECISION(Type)==64)) {                         \
           local_fp_regparm -= 1;                                \
@@ -79,8 +85,8 @@ extern "C" int ix86_regparm;
             PAttribute |= Attribute::InReg;                     \
           else                                                  \
             local_fp_regparm = 0;                               \
-      } else if (INTEGRAL_TYPE_P(Type) ||                       \
-                 POINTER_TYPE_P(Type)) {                        \
+      } else if (isa<INTEGRAL_TYPE>(Type) ||                    \
+                 isa<ACCESS_TYPE>(Type)) {                      \
           int words =                                           \
                   (Size + BITS_PER_WORD - 1) / BITS_PER_WORD;   \
           local_regparm -= words;                               \
@@ -368,6 +374,13 @@ enum x86_64_reg_class
 #define LLVM_OVERRIDE_TARGET_ARCH() \
   (TARGET_64BIT ? "x86_64" : "i386")
 
+#define LLVM_ASM_EXTENSIONS(ESCAPED_CHAR, ASM, RESULT)			\
+  else if ((ESCAPED_CHAR) == 'v') {					\
+    /* %v means to use the AVX prefix 'v' if TARGET_AVX is true. */	\
+    if (TARGET_AVX)							\
+      Result += 'v';							\
+  }
+
 /* LLVM_TARGET_INTRINSIC_LOWER - To handle builtins, we want to expand the
  * invocation into normal LLVM code.  If the target can handle the builtin, this
  * macro should call the target TreeToLLVM::TargetIntrinsicLower method and
@@ -393,10 +406,6 @@ enum x86_64_reg_class
 /* Propagate code model setting to backend */
 #define LLVM_SET_CODE_MODEL(CMModel)			\
   switch (ix86_cmodel) {				\
-  default:						\
-    sorry ("code model %<%s%> not supported yet",	\
-           ix86_cmodel_string);				\
-    break;						\
   case CM_32:						\
     CMModel = CodeModel::Default;			\
     break;						\
@@ -419,11 +428,16 @@ enum x86_64_reg_class
 
 #define LLVM_SET_MACHINE_OPTIONS(argvec)		\
   do {							\
-    if (TARGET_OMIT_LEAF_FRAME_POINTER)			\
-      argvec.push_back("--disable-non-leaf-fp-elim");	\
-							\
     if (ix86_force_align_arg_pointer)			\
       argvec.push_back("-force-align-stack");		\
   } while (0)
 
+#define LLVM_SET_TARGET_MACHINE_OPTIONS(O)		\
+  do {							\
+    if (TARGET_OMIT_LEAF_FRAME_POINTER)	{		\
+      O.NoFramePointerElim = false;			\
+      O.NoFramePointerElimNonLeaf = true;		\
+    }							\
+  } while (0)
+
 #endif /* DRAGONEGG_TARGET_H */
