@@ -1,6 +1,6 @@
 //===------- Constants.cpp - Converting and working with constants --------===//
 //
-// Copyright (C) 2011  Duncan Sands
+// Copyright (C) 2011 to 2012  Duncan Sands
 //
 // This file is part of DragonEgg.
 //
@@ -32,25 +32,31 @@
 #include "llvm/GlobalVariable.h"
 #include "llvm/LLVMContext.h"
 #include "llvm/Support/Host.h"
-#include "llvm/Support/TargetFolder.h"
 #include "llvm/Target/TargetData.h"
 
 // System headers
 #include <gmp.h>
 
 // GCC headers
+#include "auto-host.h"
+#ifndef ENABLE_BUILD_WITH_CXX
 extern "C" {
+#endif
 #include "config.h"
 // Stop GCC declaring 'getopt' as it can clash with the system's declaration.
 #undef HAVE_DECL_GETOPT
 #include "system.h"
 #include "coretypes.h"
-#include "target.h"
+#include "tm.h"
 #include "tree.h"
 
-#include "flags.h"
-#include "tm_p.h"
-}
+#if (GCC_MINOR < 7)
+#include "flags.h" // For POINTER_TYPE_OVERFLOW_UNDEFINED.
+#endif
+#include "tm_p.h"  // For CONSTANT_ALIGNMENT.
+#ifndef ENABLE_BUILD_WITH_CXX
+} // extern "C"
+#endif
 
 // Trees header.
 #include "dragonegg/Trees.h"
@@ -129,17 +135,7 @@ public:
 
   /// getBitWidth - Return the number of bits in the range.
   unsigned getBitWidth() const {
-    return R.getWidth();
-  }
-
-  /// getFirst - Return the position of the first bit in the range.
-  unsigned getFirst() const {
-    return R.getFirst();
-  }
-
-  /// getLast - Return the position of the last bit defining the range.
-  unsigned getLast() const {
-    return R.getLast();
+    return (unsigned)R.getWidth();
   }
 
   /// getRange - Return the range of bits in this slice.
@@ -176,15 +172,15 @@ BitSlice BitSlice::ExtendRange(SignedRange r, TargetFolder &Folder) const {
   if (R == r)
     return *this;
   assert(!r.empty() && "Empty ranges did not evaluate as equal?");
-  Type *ExtTy = IntegerType::get(Context, r.getWidth());
+  Type *ExtTy = IntegerType::get(Context, (unsigned)r.getWidth());
   // If the slice contains no bits then every bit of the extension is zero.
   if (empty())
     return BitSlice(r, Constant::getNullValue(ExtTy));
   // Extend the contents to the new type.
   Constant *C = Folder.CreateZExtOrBitCast(Contents, ExtTy);
   // Position the old contents correctly inside the new contents.
-  unsigned deltaFirst = R.getFirst() - r.getFirst();
-  unsigned deltaLast = r.getLast() - R.getLast();
+  unsigned deltaFirst = (unsigned)(R.getFirst() - r.getFirst());
+  unsigned deltaLast = (unsigned)(r.getLast() - R.getLast());
   if (BYTES_BIG_ENDIAN && deltaLast) {
     (void)deltaFirst; // Avoid unused variable warning.
     Constant *ShiftAmt = ConstantInt::get(C->getType(), deltaLast);
@@ -209,7 +205,7 @@ Constant *BitSlice::getBits(SignedRange r, TargetFolder &Folder) const {
   // Quick exit if the desired range matches that of the slice.
   if (R == r)
     return Contents;
-  Type *RetTy = IntegerType::get(Context, r.getWidth());
+  Type *RetTy = IntegerType::get(Context, (unsigned)r.getWidth());
   // If the slice contains no bits then every returned bit is undefined.
   if (empty())
     return UndefValue::get(RetTy);
@@ -258,8 +254,8 @@ BitSlice BitSlice::ReduceRange(SignedRange r, TargetFolder &Folder) const {
   assert(!R.empty() && "Empty ranges did not evaluate as equal?");
   // Move the least-significant bit to the correct position.
   Constant *C = Contents;
-  unsigned deltaFirst = r.getFirst() - R.getFirst();
-  unsigned deltaLast = R.getLast() - r.getLast();
+  unsigned deltaFirst = (unsigned)(r.getFirst() - R.getFirst());
+  unsigned deltaLast = (unsigned)(R.getLast() - r.getLast());
   if (BYTES_BIG_ENDIAN && deltaLast) {
     (void)deltaFirst; // Avoid unused variable warning.
     Constant *ShiftAmt = ConstantInt::get(C->getType(), deltaLast);
@@ -270,7 +266,7 @@ BitSlice BitSlice::ReduceRange(SignedRange r, TargetFolder &Folder) const {
     C = Folder.CreateLShr(C, ShiftAmt);
   }
   // Truncate to the new type.
-  Type *RedTy = IntegerType::get(Context, r.getWidth());
+  Type *RedTy = IntegerType::get(Context, (unsigned)r.getWidth());
   C = Folder.CreateTruncOrBitCast(C, RedTy);
   return BitSlice(r, C);
 }
@@ -295,7 +291,7 @@ static BitSlice ViewAsBits(Constant *C, SignedRange R, TargetFolder &Folder) {
 
   switch (Ty->getTypeID()) {
   default:
-    DieAbjectly("Unsupported type!");
+    llvm_unreachable("Unsupported type!");
   case Type::PointerTyID: {
     // Cast to an integer with the same number of bits and return that.
     IntegerType *IntTy = getTargetData().getIntPtrType(Context);
@@ -335,12 +331,15 @@ static BitSlice ViewAsBits(Constant *C, SignedRange R, TargetFolder &Folder) {
     BitSlice Bits;
     SignedRange StrideRange(0, Stride);
     for (unsigned i = FirstElt; i < LastElt; ++i) {
+      int EltOffsetInBits = i * Stride;
       // Extract the element.
       Constant *Elt = Folder.CreateExtractValue(C, i);
       // View it as a bunch of bits.
-      BitSlice EltBits = ViewAsBits(Elt, StrideRange, Folder);
+      SignedRange NeededBits = StrideRange.Meet(R.Displace(-EltOffsetInBits));
+      assert(!NeededBits.empty() && "Used element computation wrong!");
+      BitSlice EltBits = ViewAsBits(Elt, NeededBits, Folder);
       // Add to the already known bits.
-      Bits.Merge(EltBits.Displace(i * Stride), Folder);
+      Bits.Merge(EltBits.Displace(EltOffsetInBits), Folder);
     }
     return Bits;
   }
@@ -349,21 +348,26 @@ static BitSlice ViewAsBits(Constant *C, SignedRange R, TargetFolder &Folder) {
     StructType *STy = cast<StructType>(Ty);
     const StructLayout *SL = getTargetData().getStructLayout(STy);
     // Fields with indices in [FirstIdx, LastIdx) overlap the range.
-    unsigned FirstIdx = SL->getElementContainingOffset((R.getFirst()+7)/8);
-    unsigned LastIdx = 1 + SL->getElementContainingOffset((R.getLast()+6)/8);
+    unsigned FirstIdx = SL->getElementContainingOffset(R.getFirst()/8);
+    unsigned LastIdx = 1 + SL->getElementContainingOffset((R.getLast()-1)/8);
     // Visit all fields that overlap the requested range, accumulating their
     // bits in Bits.
     BitSlice Bits;
     for (unsigned i = FirstIdx; i < LastIdx; ++i) {
+      int FieldOffsetInBits = SL->getElementOffset(i) * 8;
       // Extract the field.
       Constant *Field = Folder.CreateExtractValue(C, i);
-      // View it as a bunch of bits.
+      // Only part of the field may be needed.  Compute which bits they are.
       Type *FieldTy = Field->getType();
       unsigned FieldStoreSize = getTargetData().getTypeStoreSizeInBits(FieldTy);
-      BitSlice FieldBits = ViewAsBits(Field, SignedRange(0, FieldStoreSize),
-                                      Folder);
-      // Add to the already known bits.
-      Bits.Merge(FieldBits.Displace(SL->getElementOffset(i)*8), Folder);
+      SignedRange NeededBits(0, FieldStoreSize);
+      NeededBits = NeededBits.Meet(R.Displace(-FieldOffsetInBits));
+      // View the needed part of the field as a bunch of bits.
+      if (!NeededBits.empty()) { // No field bits needed if only using padding.
+        BitSlice FieldBits = ViewAsBits(Field, NeededBits, Folder);
+        // Add to the already known bits.
+        Bits.Merge(FieldBits.Displace(FieldOffsetInBits), Folder);
+      }
     }
     return Bits;
   }
@@ -382,13 +386,16 @@ static BitSlice ViewAsBits(Constant *C, SignedRange R, TargetFolder &Folder) {
     BitSlice Bits;
     SignedRange StrideRange(0, Stride);
     for (unsigned i = FirstElt; i < LastElt; ++i) {
+      int EltOffsetInBits = i * Stride;
       // Extract the element.
       ConstantInt *Idx = ConstantInt::get(Type::getInt32Ty(Context), i);
       Constant *Elt = Folder.CreateExtractElement(C, Idx);
       // View it as a bunch of bits.
-      BitSlice EltBits = ViewAsBits(Elt, StrideRange, Folder);
+      SignedRange NeededBits = StrideRange.Meet(R.Displace(-EltOffsetInBits));
+      assert(!NeededBits.empty() && "Used element computation wrong!");
+      BitSlice EltBits = ViewAsBits(Elt, NeededBits, Folder);
       // Add to the already known bits.
-      Bits.Merge(EltBits.Displace(i * Stride), Folder);
+      Bits.Merge(EltBits.Displace(EltOffsetInBits), Folder);
     }
     return Bits;
   }
@@ -402,12 +409,20 @@ static BitSlice ViewAsBits(Constant *C, SignedRange R, TargetFolder &Folder) {
 /// value of type 'Ty' from the stored to memory location.
 static Constant *InterpretAsType(Constant *C, Type* Ty, int StartingBit,
                                  TargetFolder &Folder) {
+  // Efficient handling for some common cases.
   if (C->getType() == Ty)
     return C;
 
+  if (isa<UndefValue>(C))
+    return UndefValue::get(Ty);
+
+  if (C->isNullValue())
+    return Constant::getNullValue(Ty);
+
+  // The general case.
   switch (Ty->getTypeID()) {
   default:
-    DieAbjectly("Unsupported type!");
+    llvm_unreachable("Unsupported type!");
   case Type::IntegerTyID: {
     unsigned BitWidth = Ty->getPrimitiveSizeInBits();
     unsigned StoreSize = getTargetData().getTypeStoreSizeInBits(Ty);
@@ -501,24 +516,25 @@ static Constant *ExtractRegisterFromConstantImpl(Constant *C, tree type,
   switch (TREE_CODE(type)) {
 
   default:
-    DieAbjectly("Unknown register type!", type);
+    debug_tree(type);
+    llvm_unreachable("Unknown register type!");
 
   case BOOLEAN_TYPE:
   case ENUMERAL_TYPE:
   case INTEGER_TYPE: {
-    // For integral types, extract an integer with size equal to the type size,
+    // For integral types, extract an integer with size equal to the mode size,
     // then truncate down to the precision.  For example, when extracting a bool
     // this probably first loads out an i8 or i32 which is then truncated to i1.
     // This roundabout approach means we get the right result on both little and
     // big endian machines.
-    uint64_t Size = getInt64(TYPE_SIZE(type), true);
+    unsigned Size = GET_MODE_BITSIZE(TYPE_MODE(type));
     Type *MemTy = IntegerType::get(Context, Size);
     C = InterpretAsType(C, MemTy, StartingBit, Folder);
     return Folder.CreateTruncOrBitCast(C, getRegType(type));
   }
 
   case COMPLEX_TYPE: {
-    tree elt_type = TREE_TYPE(type);
+    tree elt_type = main_type(type);
     unsigned Stride = GET_MODE_BITSIZE(TYPE_MODE(elt_type));
     Constant *Vals[2] = {
       ExtractRegisterFromConstantImpl(C, elt_type, StartingBit, Folder),
@@ -538,7 +554,7 @@ static Constant *ExtractRegisterFromConstantImpl(Constant *C, tree type,
     return InterpretAsType(C, getRegType(type), StartingBit, Folder);
 
   case VECTOR_TYPE: {
-    tree elt_type = TREE_TYPE(type);
+    tree elt_type = main_type(type);
     unsigned NumElts = TYPE_VECTOR_SUBPARTS(type);
     unsigned Stride = GET_MODE_BITSIZE(TYPE_MODE(elt_type));
     SmallVector<Constant*, 16> Vals(NumElts);
@@ -576,7 +592,7 @@ Constant *ExtractRegisterFromConstant(Constant *C, tree type, int StartingByte)
 /// register type.
 static Constant *getAsRegister(tree exp, TargetFolder &Folder) {
   Constant *C = ConvertInitializerImpl(exp, Folder);
-  return ExtractRegisterFromConstantImpl(C, TREE_TYPE(exp), 0, Folder);
+  return ExtractRegisterFromConstantImpl(C, main_type(exp), 0, Folder);
 }
 
 /// RepresentAsMemory - Turn a constant of in-register type (corresponding
@@ -592,18 +608,18 @@ static Constant *RepresentAsMemory(Constant *C, tree type,
   switch (TREE_CODE(type)) {
 
   default:
-    DieAbjectly("Unknown register type!", type);
+    debug_tree(type);
+    llvm_unreachable("Unknown register type!");
 
   case BOOLEAN_TYPE:
   case ENUMERAL_TYPE:
   case INTEGER_TYPE: {
-    // For integral types extend to an integer with size equal to the type size.
+    // For integral types extend to an integer with size equal to the mode size.
     // For example, when inserting a bool this probably extends it to an i8 or
     // to an i32.  This approach means we get the right result on both little
     // and big endian machines.
-    uint64_t Size = getInt64(TYPE_SIZE(type), true);
+    unsigned Size = GET_MODE_BITSIZE(TYPE_MODE(type));
     Type *MemTy = IntegerType::get(Context, Size);
-    // We can extend in any way, but get nicer IR by respecting signedness.
     bool isSigned = !TYPE_UNSIGNED(type);
     Result = isSigned ? Folder.CreateSExtOrBitCast(C, MemTy) :
       Folder.CreateZExtOrBitCast(C, MemTy);
@@ -611,10 +627,9 @@ static Constant *RepresentAsMemory(Constant *C, tree type,
   }
 
   case COMPLEX_TYPE: {
-    tree elt_type = TREE_TYPE(type);
-    unsigned Idx[2] = {0, 1};
-    Constant *Real = Folder.CreateExtractValue(C, Idx[0]);
-    Constant *Imag = Folder.CreateExtractValue(C, Idx[1]);
+    tree elt_type = main_type(type);
+    Constant *Real = Folder.CreateExtractValue(C, 0);
+    Constant *Imag = Folder.CreateExtractValue(C, 1);
     Real = RepresentAsMemory(Real, elt_type, Folder);
     Imag = RepresentAsMemory(Imag, elt_type, Folder);
     Constant *Vals[2] = { Real, Imag };
@@ -637,7 +652,7 @@ static Constant *RepresentAsMemory(Constant *C, tree type,
     break;
 
   case VECTOR_TYPE: {
-    tree elt_type = TREE_TYPE(type);
+    tree elt_type = main_type(type);
     unsigned NumElts = TYPE_VECTOR_SUBPARTS(type);
     std::vector<Constant*> Vals(NumElts);
     for (unsigned i = 0; i != NumElts; ++i) {
@@ -665,24 +680,29 @@ static Constant *RepresentAsMemory(Constant *C, tree type,
 
 /// ConvertInitializerWithCast - Convert the initial value for a global variable
 /// to an equivalent LLVM constant then cast to the given type if both the type
-/// and the initializer are scalar.  This is convenient for making explicit the
-/// implicit scalar casts that GCC allows in "assignments" such as initializing
-/// a record field.
+/// and the initializer are scalar, or if the initializer's type only differs in
+/// trivial ways from the given type.  This is convenient for avoiding confusing
+/// and pointless type changes in the IR, and for making explicit the implicit
+/// scalar casts that GCC allows in "assignments" such as initializing a record
+/// field.
 static Constant *ConvertInitializerWithCast(tree exp, tree type,
                                             TargetFolder &Folder) {
-  // Convert the initializer.
+  // Convert the initializer.  Note that the type of the returned value may be
+  // pretty much anything.
   Constant *C = ConvertInitializerImpl(exp, Folder);
 
-  // If no cast is needed, or it would not be a scalar cast, then just return
-  // the initializer as is.
-  if (type == TREE_TYPE(exp) || AGGREGATE_TYPE_P(TREE_TYPE(exp)) ||
-      AGGREGATE_TYPE_P(type))
+  // If casting to or from an aggregate then just return the initializer as is.
+  // If the types differ then this is probably something like a struct ending in
+  // a flexible array being initialized with a struct ending in an array of some
+  // definite size.
+  if (isa<AGGREGATE_TYPE>(type) || isa<AGGREGATE_TYPE>(TREE_TYPE(exp)))
     return C;
-  // No cast is needed if the LLVM types are the same.  This occurs often since
-  // many different GCC types usually map to the same LLVM type.
-  Type *SrcTy = getRegType(TREE_TYPE(exp));
+
+  // Scalar to scalar cast.  This is where the implicit scalar casts that GCC
+  // permits are made explicit.
   Type *DestTy = getRegType(type);
-  if (SrcTy == DestTy)
+  if (C->getType() == DestTy)
+    // No cast is needed if the type is already correct.
     return C;
 
   // Ensure that the initializer has a sensible type.  Note that it would be
@@ -690,7 +710,7 @@ static Constant *ConvertInitializerWithCast(tree exp, tree type,
   // that would not perform a value extension (adding extra zeros or sign bits
   // when casting to a larger integer type for example): any extra bits would
   // wrongly get an undefined value instead.
-  C = ExtractRegisterFromConstantImpl(C, TREE_TYPE(exp), 0, Folder);
+  C = ExtractRegisterFromConstantImpl(C, main_type(exp), 0, Folder);
 
   // Cast to the desired type.
   bool SrcIsSigned = !TYPE_UNSIGNED(TREE_TYPE(exp));
@@ -705,17 +725,16 @@ static Constant *ConvertInitializerWithCast(tree exp, tree type,
 /// ConvertCST - Return the given simple constant as an array of bytes.  For the
 /// moment only INTEGER_CST, REAL_CST, COMPLEX_CST and VECTOR_CST are supported.
 static Constant *ConvertCST(tree exp, TargetFolder &) {
-  const tree type = TREE_TYPE(exp);
+  const tree type = main_type(exp);
   unsigned SizeInChars = (TREE_INT_CST_LOW(TYPE_SIZE(type)) + CHAR_BIT - 1) /
     CHAR_BIT;
   // Encode the constant in Buffer in target format.
-  std::vector<unsigned char> Buffer(SizeInChars);
+  SmallVector<uint8_t, 16> Buffer(SizeInChars);
   unsigned CharsWritten = native_encode_expr(exp, &Buffer[0], SizeInChars);
   assert(CharsWritten == SizeInChars && "Failed to fully encode expression!");
   (void)CharsWritten; // Avoid unused variable warning when assertions disabled.
   // Turn it into an LLVM byte array.
-  return ConstantArray::get(Context, StringRef((char *)&Buffer[0], SizeInChars),
-                            /*AddNull*/false);
+  return ConstantDataArray::get(Context, Buffer);
 }
 
 static Constant *ConvertSTRING_CST(tree exp, TargetFolder &) {
@@ -761,17 +780,17 @@ static Constant *ConvertSTRING_CST(tree exp, TargetFolder &) {
                                         ByteSwap_32(InStr[i])));
     }
   } else {
-    assert(0 && "Unknown character type!");
+    llvm_unreachable("Unknown character type!");
   }
 
   unsigned LenInElts = Len /
-          TREE_INT_CST_LOW(TYPE_SIZE_UNIT(TREE_TYPE(TREE_TYPE(exp))));
+          TREE_INT_CST_LOW(TYPE_SIZE_UNIT(main_type(main_type(exp))));
   unsigned ConstantSize = StrTy->getNumElements();
 
   if (LenInElts != ConstantSize) {
     // If this is a variable sized array type, set the length to LenInElts.
     if (ConstantSize == 0) {
-      tree Domain = TYPE_DOMAIN(TREE_TYPE(exp));
+      tree Domain = TYPE_DOMAIN(main_type(exp));
       if (!Domain || !TYPE_MAX_VALUE(Domain)) {
         ConstantSize = LenInElts;
         StrTy = ArrayType::get(ElTy, LenInElts);
@@ -799,10 +818,10 @@ static Constant *ConvertADDR_EXPR(tree exp, TargetFolder &Folder) {
 static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
   const TargetData &TD = getTargetData();
 
-  tree init_type = TREE_TYPE(exp);
+  tree init_type = main_type(exp);
   Type *InitTy = ConvertType(init_type);
 
-  tree elt_type = TREE_TYPE(init_type);
+  tree elt_type = main_type(init_type);
   Type *EltTy = ConvertType(elt_type);
 
   // Check that the element type has a known, constant size.
@@ -816,7 +835,7 @@ static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
   // Resize to the number of array elements if known.  This ensures that every
   // element will be at least default initialized even if no initial value is
   // given for it.
-  uint64_t TypeElts = TREE_CODE(init_type) == ARRAY_TYPE ?
+  uint64_t TypeElts = isa<ARRAY_TYPE>(init_type) ?
     ArrayLengthOf(init_type) : TYPE_VECTOR_SUBPARTS(init_type);
   if (TypeElts != NO_LENGTH)
     Elts.resize(TypeElts);
@@ -824,7 +843,7 @@ static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
   // If GCC indices into the array need adjusting to make them zero indexed then
   // record here the value to subtract off.
   tree lower_bnd = NULL_TREE;
-  if (TREE_CODE(init_type) == ARRAY_TYPE && TYPE_DOMAIN(init_type) &&
+  if (isa<ARRAY_TYPE>(init_type) && TYPE_DOMAIN(init_type) &&
       !integer_zerop(TYPE_MIN_VALUE(TYPE_DOMAIN(init_type))))
     lower_bnd = TYPE_MIN_VALUE(TYPE_DOMAIN(init_type));
 
@@ -842,11 +861,11 @@ static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
       unsigned PadBits = EltSize - ValSize;
       assert(PadBits % BITS_PER_UNIT == 0 && "Non-unit type size?");
       unsigned Units = PadBits / BITS_PER_UNIT;
-      Constant *Elts[] = {
+      Constant *PaddedElt[] = {
         Val, UndefValue::get(GetUnitType(Context, Units))
       };
-      
-      Val = ConstantStruct::getAnon(Elts);
+
+      Val = ConstantStruct::getAnon(PaddedElt);
     }
 
     // Get the index position of the element within the array.  Note that this
@@ -857,14 +876,14 @@ static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
     unsigned FirstIndex, LastIndex;
     if (!index) {
       LastIndex = FirstIndex = NextIndex;
-    } else if (TREE_CODE(index) == RANGE_EXPR) {
+    } else if (isa<RANGE_EXPR>(index)) {
       tree first = TREE_OPERAND(index, 0);
       tree last  = TREE_OPERAND(index, 1);
 
       // Subtract off the lower bound if any to ensure indices start from zero.
       if (lower_bnd != NULL_TREE) {
-        first = fold_build2(MINUS_EXPR, TREE_TYPE(first), first, lower_bnd);
-        last = fold_build2(MINUS_EXPR, TREE_TYPE(last), last, lower_bnd);
+        first = fold_build2(MINUS_EXPR, main_type(first), first, lower_bnd);
+        last = fold_build2(MINUS_EXPR, main_type(last), last, lower_bnd);
       }
 
       assert(host_integerp(first, 1) && host_integerp(last, 1) &&
@@ -874,7 +893,7 @@ static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
     } else {
       // Subtract off the lower bound if any to ensure indices start from zero.
       if (lower_bnd != NULL_TREE)
-        index = fold_build2(MINUS_EXPR, TREE_TYPE(index), index, lower_bnd);
+        index = fold_build2(MINUS_EXPR, main_type(index), index, lower_bnd);
       assert(host_integerp(index, 1));
       FirstIndex = tree_low_cst(index, 1);
       LastIndex = FirstIndex;
@@ -905,20 +924,15 @@ static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
   // return a struct instead of an array.  This can occur in cases where we have
   // an array of unions, and the various unions had different parts initialized.
   // While there, compute the maximum element alignment.
-  bool UseStruct = false;
+  bool isHomogeneous = true;
   Type *ActualEltTy = Elts[0]->getType();
   unsigned MaxAlign = TD.getABITypeAlignment(ActualEltTy);
   for (unsigned i = 1; i != NumElts; ++i)
     if (Elts[i]->getType() != ActualEltTy) {
       MaxAlign = std::max(TD.getABITypeAlignment(Elts[i]->getType()), MaxAlign);
-      UseStruct = true;
+      isHomogeneous = false;
     }
 
-  // If any elements are more aligned than the GCC type then we need to return a
-  // packed struct.  This can happen if the user forced a small alignment on the
-  // array type.
-  bool Pack = MaxAlign * 8 > TYPE_ALIGN(TREE_TYPE(exp));
-
   // We guarantee that initializers are always at least as big as the LLVM type
   // for the initializer.  If needed, append padding to ensure this.
   uint64_t TypeSize = TD.getTypeAllocSizeInBits(InitTy);
@@ -927,17 +941,45 @@ static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
     assert(PadBits % BITS_PER_UNIT == 0 && "Non-unit type size?");
     unsigned Units = PadBits / BITS_PER_UNIT;
     Elts.push_back(UndefValue::get(GetUnitType(Context, Units)));
-    UseStruct = true;
+    isHomogeneous = false;
   }
 
+  // If any elements are more aligned than the GCC type then we need to return a
+  // packed struct.  This can happen if the user forced a small alignment on the
+  // array type.
+  if (MaxAlign * 8 > TYPE_ALIGN(main_type(exp)))
+    return ConstantStruct::getAnon(Context, Elts, /*Packed*/true);
+
   // Return as a struct if the contents are not homogeneous.
-  if (UseStruct || Pack)
-    return ConstantStruct::getAnon(Context, Elts, Pack);
+  if (!isHomogeneous) {
+    std::vector<Constant*> StructElts;
+    unsigned First = 0, E = Elts.size();
+    while (First < E) {
+      // Find the maximal value of Last s.t. all elements in the range
+      // [First, Last) have the same type.
+      Type *Ty = Elts[First]->getType();
+      unsigned Last = First + 1;
+      for (; Last != E; ++Last)
+        if (Elts[Last]->getType() != Ty)
+          break;
+      unsigned NumSameType = Last - First;
+      Constant *StructElt;
+      if (NumSameType == 1)
+        StructElt = Elts[First];
+      else
+        StructElt = ConstantArray::get(ArrayType::get(Ty, NumSameType),
+                                       ArrayRef<Constant*>(&Elts[First],
+                                                           NumSameType));
+      StructElts.push_back(StructElt);
+      First = Last;
+    }
+    return ConstantStruct::getAnon(Context, StructElts);
+  }
 
   // Make the IR more pleasant by returning as a vector if the GCC type was a
   // vector.  However this is only correct if the initial values had the same
   // type as the vector element type, rather than some random other type.
-  if (TREE_CODE(init_type) == VECTOR_TYPE && ActualEltTy == EltTy) {
+  if (isa<VECTOR_TYPE>(init_type) && ActualEltTy == EltTy) {
     // If this is a vector of pointers, convert it to a vector of integers.
     if (isa<PointerType>(EltTy)) {
       IntegerType *IntPtrTy = getTargetData().getIntPtrType(Context);
@@ -955,13 +997,13 @@ static Constant *ConvertArrayCONSTRUCTOR(tree exp, TargetFolder &Folder) {
 namespace {
 
 class FieldContents {
+  TargetFolder &Folder;
   SignedRange R; // The range of bits occupied by the constant.
   Constant *C;   // The constant.  May be null if the range is empty.
   int Starts;    // The first bit of the constant is positioned at this offset.
-  TargetFolder &Folder;
 
   FieldContents(SignedRange r, Constant *c, int starts, TargetFolder &folder)
-    : R(r), C(c), Starts(starts), Folder(folder) {
+    : Folder(folder), R(r), C(c), Starts(starts) {
     assert((R.empty() || C) && "Need constant when range not empty!");
   }
 
@@ -1093,7 +1135,9 @@ static Constant *ConvertRecordCONSTRUCTOR(tree exp, TargetFolder &Folder) {
   // and probably wrong on big-endian machines.
   IntervalList<FieldContents, int, 8> Layout;
   const TargetData &TD = getTargetData();
-  uint64_t TypeSize = TD.getTypeAllocSizeInBits(ConvertType(TREE_TYPE(exp)));
+  tree type = main_type(exp);
+  Type *Ty = ConvertType(type);
+  uint64_t TypeSize = TD.getTypeAllocSizeInBits(Ty);
 
   // Ensure that fields without an initial value are default initialized by
   // explicitly setting the starting value for all fields to be zero.  If an
@@ -1102,17 +1146,12 @@ static Constant *ConvertRecordCONSTRUCTOR(tree exp, TargetFolder &Folder) {
   if (flag_default_initialize_globals) {
     // Record all interesting fields so they can easily be visited backwards.
     SmallVector<tree, 16> Fields;
-    for (tree field = TYPE_FIELDS(TREE_TYPE(exp)); field;
-         field = TREE_CHAIN(field)) {
-      assert(TREE_CODE(field) == FIELD_DECL && "Lang data not freed?");
+    for (tree field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
+      if (!isa<FIELD_DECL>(field)) continue;
       // Ignore fields with variable or unknown position since they cannot be
       // default initialized.
       if (!OffsetIsLLVMCompatible(field))
         continue;
-      // Skip fields that are known not to be present.
-      if (TREE_CODE(TREE_TYPE(exp)) == QUAL_UNION_TYPE &&
-          integer_zerop(DECL_QUALIFIER(field)))
-        continue; 
       Fields.push_back(field);
     }
 
@@ -1158,23 +1197,23 @@ static Constant *ConvertRecordCONSTRUCTOR(tree exp, TargetFolder &Folder) {
   // occupied by the field to that value.
   unsigned HOST_WIDE_INT ix;
   tree field, next_field, value;
-  next_field = TYPE_FIELDS(TREE_TYPE(exp));
+  next_field = TYPE_FIELDS(type);
   FOR_EACH_CONSTRUCTOR_ELT(CONSTRUCTOR_ELTS(exp), ix, field, value) {
     if (!field) {
       // Move on to the next FIELD_DECL, skipping contained methods, types etc.
       field = next_field;
       while (1) {
         assert(field && "Fell off end of record!");
-        if (TREE_CODE(field) == FIELD_DECL) break;
+        if (isa<FIELD_DECL>(field)) break;
         field = TREE_CHAIN(field);
       }
     }
     next_field = TREE_CHAIN(field);
 
-    assert(TREE_CODE(field) == FIELD_DECL && "Initial value not for a field!");
+    assert(isa<FIELD_DECL>(field) && "Initial value not for a field!");
     assert(OffsetIsLLVMCompatible(field) && "Field position not known!");
     // Turn the initial value for this field into an LLVM constant.
-    Constant *Init = ConvertInitializerWithCast(value, TREE_TYPE(field),
+    Constant *Init = ConvertInitializerWithCast(value, main_type(field),
                                                 Folder);
     // Work out the range of bits occupied by the field.
     uint64_t FirstBit = getFieldOffsetInBits(field);
@@ -1199,7 +1238,7 @@ static Constant *ConvertRecordCONSTRUCTOR(tree exp, TargetFolder &Folder) {
   // type then return a packed struct instead.  If a field's alignment would
   // make it start after its desired position then also use a packed struct.
   bool Pack = false;
-  unsigned MaxAlign = TYPE_ALIGN(TREE_TYPE(exp));
+  unsigned MaxAlign = TYPE_ALIGN(type);
   for (unsigned i = 0, e = Layout.getNumIntervals(); i != e; ++i) {
     FieldContents F = Layout.getInterval(i);
     unsigned First = F.getRange().getFirst();
@@ -1258,7 +1297,33 @@ static Constant *ConvertRecordCONSTRUCTOR(tree exp, TargetFolder &Folder) {
     Elts.push_back(UndefValue::get(GetUnitType(Context, Units)));
   }
 
-  // Okay, we're done, return the computed elements.
+  // Okay, we're done.  Return the computed elements as a constant with the type
+  // of exp if possible.
+  if (StructType *STy = dyn_cast<StructType>(Ty))
+    if (STy->isPacked() == Pack && STy->getNumElements() == Elts.size()) {
+      bool EltTypesMatch = true;
+      for (unsigned i = 0, e = Elts.size(); i != e; ++i) {
+        Type *EltTy = Elts[i]->getType();
+        Type *FieldTy = STy->getElementType(i);
+        if (EltTy == FieldTy)
+          continue;
+        // When a recursive record type is converted, some of its pointer fields
+        // may be converted to the artifical type {}* to break the recursion. As
+        // type converting the field directly gives the proper pointer type, the
+        // result is a mismatch between the field and element types.  Fix it up.
+        if (EltTy->isPointerTy() && FieldTy->isPointerTy()) {
+          Elts[i] = Folder.CreateBitCast(Elts[i], FieldTy);
+          continue;
+        }
+        // Too hard, just give up.
+        EltTypesMatch = false;
+        break;
+      }
+      if (EltTypesMatch)
+        return ConstantStruct::get(STy, Elts);
+    }
+
+  // Otherwise return the computed elements as an anonymous struct.
   return ConstantStruct::getAnon(Context, Elts, Pack);
 }
 
@@ -1271,7 +1336,8 @@ static Constant *ConvertCONSTRUCTOR(tree exp, TargetFolder &Folder) {
 
   switch (TREE_CODE(TREE_TYPE(exp))) {
   default:
-    DieAbjectly("Unknown constructor!", exp);
+    debug_tree(exp);
+    llvm_unreachable("Unknown constructor!");
   case VECTOR_TYPE:
   case ARRAY_TYPE:  return ConvertArrayCONSTRUCTOR(exp, Folder);
   case QUAL_UNION_TYPE:
@@ -1283,13 +1349,13 @@ static Constant *ConvertCONSTRUCTOR(tree exp, TargetFolder &Folder) {
 static Constant *ConvertMINUS_EXPR(tree exp, TargetFolder &Folder) {
   Constant *LHS = getAsRegister(TREE_OPERAND(exp, 0), Folder);
   Constant *RHS = getAsRegister(TREE_OPERAND(exp, 1), Folder);
-  return RepresentAsMemory(Folder.CreateSub(LHS, RHS), TREE_TYPE(exp), Folder);
+  return RepresentAsMemory(Folder.CreateSub(LHS, RHS), main_type(exp), Folder);
 }
 
 static Constant *ConvertPLUS_EXPR(tree exp, TargetFolder &Folder) {
   Constant *LHS = getAsRegister(TREE_OPERAND(exp, 0), Folder);
   Constant *RHS = getAsRegister(TREE_OPERAND(exp, 1), Folder);
-  return RepresentAsMemory(Folder.CreateAdd(LHS, RHS), TREE_TYPE(exp), Folder);
+  return RepresentAsMemory(Folder.CreateAdd(LHS, RHS), main_type(exp), Folder);
 }
 
 static Constant *ConvertPOINTER_PLUS_EXPR(tree exp, TargetFolder &Folder) {
@@ -1305,7 +1371,7 @@ static Constant *ConvertPOINTER_PLUS_EXPR(tree exp, TargetFolder &Folder) {
   // The result may be of a different pointer type.
   Result = Folder.CreateBitCast(Result, getRegType(TREE_TYPE(exp)));
 
-  return RepresentAsMemory(Result, TREE_TYPE(exp), Folder);
+  return RepresentAsMemory(Result, main_type(exp), Folder);
 }
 
 static Constant *ConvertVIEW_CONVERT_EXPR(tree exp, TargetFolder &Folder) {
@@ -1315,7 +1381,8 @@ static Constant *ConvertVIEW_CONVERT_EXPR(tree exp, TargetFolder &Folder) {
 
 /// ConvertInitializerImpl - Implementation of ConvertInitializer.
 static Constant *ConvertInitializerImpl(tree exp, TargetFolder &Folder) {
-  assert(!HAS_RTL_P(exp) && "Cache collision with decl_llvm!");
+  assert(!isa<CONST_DECL>(exp) && !HAS_RTL_P(exp) &&
+         "Cache collision with decl_llvm!");
 
   // If we already converted the initializer then return the cached copy.
   if (Constant *C = cast_or_null<Constant>(getCachedValue(exp)))
@@ -1324,7 +1391,8 @@ static Constant *ConvertInitializerImpl(tree exp, TargetFolder &Folder) {
   Constant *Init;
   switch (TREE_CODE(exp)) {
   default:
-    DieAbjectly("Unknown constant to convert!", exp);
+    debug_tree(exp);
+    llvm_unreachable("Unknown constant to convert!");
   case COMPLEX_CST:
   case INTEGER_CST:
   case REAL_CST:
@@ -1342,7 +1410,7 @@ static Constant *ConvertInitializerImpl(tree exp, TargetFolder &Folder) {
     break;
   case CONVERT_EXPR:
   case NOP_EXPR:
-    Init = ConvertInitializerWithCast(TREE_OPERAND(exp, 0), TREE_TYPE(exp),
+    Init = ConvertInitializerWithCast(TREE_OPERAND(exp, 0), main_type(exp),
                                       Folder);
     break;
   case MINUS_EXPR:
@@ -1359,9 +1427,9 @@ static Constant *ConvertInitializerImpl(tree exp, TargetFolder &Folder) {
     break;
   }
 
-  // Make the IR easier to read by converting the bunch of bytes returned by
-  // ConvertCST into a less surprising type when it is safe to do so.
-  if (!AGGREGATE_TYPE_P(TREE_TYPE(exp)))
+  // Make the IR easier to read by returning a constant of the expected type if
+  // it is safe and efficient to do so.
+  if (!isa<AGGREGATE_TYPE>(TREE_TYPE(exp)))
     Init = InterpretAsType(Init, ConvertType(TREE_TYPE(exp)), 0, Folder);
 
 #ifndef NDEBUG
@@ -1373,14 +1441,16 @@ static Constant *ConvertInitializerImpl(tree exp, TargetFolder &Folder) {
   if (Ty->isSized()) {
     uint64_t InitSize = getTargetData().getTypeAllocSizeInBits(Init->getType());
     uint64_t TypeSize = getTargetData().getTypeAllocSizeInBits(Ty);
-    if (InitSize < TypeSize)
-      DieAbjectly("Constant too small for type!", exp);
-    if (isInt64(TREE_TYPE(exp), true) && InitSize != TypeSize)
-      DieAbjectly("Constant too big for type!", exp);
+    if (InitSize < TypeSize) {
+      debug_tree(exp);
+      llvm_unreachable("Constant too small for type!");
+    }
   }
   if (getTargetData().getABITypeAlignment(Init->getType()) * 8 >
-      TYPE_ALIGN(TREE_TYPE(exp)))
-    DieAbjectly("Constant over aligned!", exp);
+      TYPE_ALIGN(main_type(exp))) {
+    debug_tree(exp);
+    llvm_unreachable("Constant over aligned!");
+  }
 #endif
 
   // Cache the result of converting the initializer since the same tree is often
@@ -1423,7 +1493,7 @@ static Constant *AddressOfCST(tree exp, TargetFolder &Folder) {
   // Create a new global variable.
   Slot = new GlobalVariable(*TheModule, Init->getType(), true,
                             GlobalVariable::LinkerPrivateLinkage, Init, ".cst");
-  unsigned align = TYPE_ALIGN(TREE_TYPE(exp));
+  unsigned align = TYPE_ALIGN(main_type(exp));
 #ifdef CONSTANT_ALIGNMENT
   align = CONSTANT_ALIGNMENT(exp, align);
 #endif
@@ -1440,11 +1510,11 @@ static Constant *AddressOfCST(tree exp, TargetFolder &Folder) {
 static Constant *AddressOfARRAY_REF(tree exp, TargetFolder &Folder) {
   tree array = TREE_OPERAND(exp, 0);
   tree index = TREE_OPERAND(exp, 1);
-  tree index_type = TREE_TYPE(index);
-  assert(TREE_CODE(TREE_TYPE(array)) == ARRAY_TYPE && "Unknown ARRAY_REF!");
+  tree index_type = main_type(index);
+  assert(isa<ARRAY_TYPE>(TREE_TYPE(array)) && "Unknown ARRAY_REF!");
 
   // Check for variable sized reference.
-  assert(isSizeCompatible(TREE_TYPE(TREE_TYPE(array))) &&
+  assert(isSizeCompatible(main_type(main_type(array))) &&
          "Global with variable size?");
 
   // Get the index into the array as an LLVM integer constant.
@@ -1462,7 +1532,7 @@ static Constant *AddressOfARRAY_REF(tree exp, TargetFolder &Folder) {
   // Avoid any assumptions about how the array type is represented in LLVM by
   // doing the GEP on a pointer to the first array element.
   Constant *ArrayAddr = AddressOfImpl(array, Folder);
-  Type *EltTy = ConvertType(TREE_TYPE(TREE_TYPE(array)));
+  Type *EltTy = ConvertType(main_type(main_type(array)));
   ArrayAddr = Folder.CreateBitCast(ArrayAddr, EltTy->getPointerTo());
 
   return POINTER_TYPE_OVERFLOW_UNDEFINED ?
@@ -1530,7 +1600,7 @@ static Constant *AddressOfLABEL_DECL(tree exp, TargetFolder &) {
 
   // Figure out which function this is for, verify it's the one we're compiling.
   if (DECL_CONTEXT(exp)) {
-    assert(TREE_CODE(DECL_CONTEXT(exp)) == FUNCTION_DECL &&
+    assert(isa<FUNCTION_DECL>(DECL_CONTEXT(exp)) &&
            "Address of label in nested function?");
     assert(TheTreeToLLVM->getFUNCTION_DECL() == DECL_CONTEXT(exp) &&
            "Taking the address of a label that isn't in the current fn!?");
@@ -1539,13 +1609,31 @@ static Constant *AddressOfLABEL_DECL(tree exp, TargetFolder &) {
   return TheTreeToLLVM->AddressOfLABEL_DECL(exp);
 }
 
+#if (GCC_MINOR > 5)
+/// AddressOfMEM_REF - Return the address of a memory reference.
+static Constant *AddressOfMEM_REF(tree exp, TargetFolder &Folder) {
+  // The address is the first operand offset in bytes by the second.
+  Constant *Addr = getAsRegister(TREE_OPERAND(exp, 0), Folder);
+  if (integer_zerop(TREE_OPERAND(exp, 1)))
+    return Addr;
+
+  // Convert to a byte pointer and displace by the offset.
+  Addr = Folder.CreateBitCast(Addr, GetUnitPointerType(Context));
+  APInt Delta = getIntegerValue(TREE_OPERAND(exp, 1));
+  Constant *Offset = ConstantInt::get(Context, Delta);
+  // The address is always inside the referenced object, so "inbounds".
+  return Folder.CreateInBoundsGetElementPtr(Addr, Offset);
+}
+#endif
+
 /// AddressOfImpl - Implementation of AddressOf.
 static Constant *AddressOfImpl(tree exp, TargetFolder &Folder) {
   Constant *Addr;
 
   switch (TREE_CODE(exp)) {
   default:
-    DieAbjectly("Unknown constant to take the address of!", exp);
+    debug_tree(exp);
+    llvm_unreachable("Unknown constant to take the address of!");
   case COMPLEX_CST:
   case FIXED_CST:
   case INTEGER_CST:
@@ -1578,12 +1666,17 @@ static Constant *AddressOfImpl(tree exp, TargetFolder &Folder) {
   case LABEL_DECL:
     Addr = AddressOfLABEL_DECL(exp, Folder);
     break;
+#if (GCC_MINOR > 5)
+  case MEM_REF:
+    Addr = AddressOfMEM_REF(exp, Folder);
+    break;
+#endif
   }
 
   // Ensure that the address has the expected type.  It is simpler to do this
   // once here rather than in every AddressOf helper.
   Type *Ty;
-  if (VOID_TYPE_P(TREE_TYPE(exp)))
+  if (isa<VOID_TYPE>(TREE_TYPE(exp)))
     Ty = GetUnitPointerType(Context); // void* -> i8*.
   else
     Ty = ConvertType(TREE_TYPE(exp))->getPointerTo();
