@@ -1,7 +1,6 @@
 //===-------------- Debug.cpp - Debug information gathering ---------------===//
 //
-// Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011  Jim Laskey,
-// Duncan Sands et al.
+// Copyright (C) 2005 to 2012  Jim Laskey, Duncan Sands et al.
 //
 // This file is part of DragonEgg.
 //
@@ -26,26 +25,30 @@
 
 // LLVM headers
 #include "llvm/Module.h"
-#include "llvm/ADT/STLExtras.h"
 
 // System headers
 #include <gmp.h>
 
 // GCC headers
+#include "auto-host.h"
+#ifndef ENABLE_BUILD_WITH_CXX
 extern "C" {
+#endif
 #include "config.h"
 // Stop GCC declaring 'getopt' as it can clash with the system's declaration.
 #undef HAVE_DECL_GETOPT
 #include "system.h"
 #include "coretypes.h"
-#include "target.h"
+#include "tm.h"
 #include "tree.h"
 
 #include "flags.h"
 #include "langhooks.h"
 #include "toplev.h"
 #include "version.h"
-}
+#ifndef ENABLE_BUILD_WITH_CXX
+} // extern "C"
+#endif
 
 // Trees header.
 #include "dragonegg/Trees.h"
@@ -87,9 +90,9 @@ static void DirectoryAndFile(const std::string &FullPath,
 /// NodeSizeInBits - Returns the size in bits stored in a tree node regardless
 /// of whether the node is a TYPE or DECL.
 static uint64_t NodeSizeInBits(tree Node) {
-  if (TREE_CODE(Node) == ERROR_MARK) {
+  if (isa<ERROR_MARK>(Node)) {
     return BITS_PER_WORD;
-  } else if (TYPE_P(Node)) {
+  } else if (isa<TYPE>(Node)) {
     if (TYPE_SIZE(Node) == NULL_TREE)
       return 0;
     else if (isInt64(TYPE_SIZE(Node), true))
@@ -111,8 +114,8 @@ static uint64_t NodeSizeInBits(tree Node) {
 /// NodeAlignInBits - Returns the alignment in bits stored in a tree node
 /// regardless of whether the node is a TYPE or DECL.
 static uint64_t NodeAlignInBits(tree Node) {
-  if (TREE_CODE(Node) == ERROR_MARK) return BITS_PER_WORD;
-  if (TYPE_P(Node)) return TYPE_ALIGN(Node);
+  if (isa<ERROR_MARK>(Node)) return BITS_PER_WORD;
+  if (isa<TYPE>(Node)) return TYPE_ALIGN(Node);
   if (DECL_P(Node)) return DECL_ALIGN(Node);
   return BITS_PER_WORD;
 }
@@ -120,7 +123,7 @@ static uint64_t NodeAlignInBits(tree Node) {
 /// FieldType - Returns the type node of a structure member field.
 ///
 static tree FieldType(tree Field) {
-  if (TREE_CODE (Field) == ERROR_MARK) return integer_type_node;
+  if (isa<ERROR_MARK>(Field)) return integer_type_node;
   return DECL_BIT_FIELD_TYPE(Field) ?
     DECL_BIT_FIELD_TYPE(Field) : TREE_TYPE (Field);
 }
@@ -132,14 +135,14 @@ static StringRef GetNodeName(tree Node) {
 
   if (DECL_P(Node)) {
     Name = DECL_NAME(Node);
-  } else if (TYPE_P(Node)) {
+  } else if (isa<TYPE>(Node)) {
     Name = TYPE_NAME(Node);
   }
 
   if (Name) {
-    if (TREE_CODE(Name) == IDENTIFIER_NODE) {
+    if (isa<IDENTIFIER_NODE>(Name)) {
       return IDENTIFIER_POINTER(Name);
-    } else if (TREE_CODE(Name) == TYPE_DECL && DECL_NAME(Name) &&
+    } else if (isa<TYPE_DECL>(Name) && DECL_NAME(Name) &&
                !DECL_IGNORED_P(Name)) {
       return StringRef(IDENTIFIER_POINTER(DECL_NAME(Name)));
     }
@@ -161,12 +164,12 @@ static expanded_location GetNodeLocation(tree Node, bool UseStub = true) {
 
   if (DECL_P(Node)) {
     Name = DECL_NAME(Node);
-  } else if (TYPE_P(Node)) {
+  } else if (isa<TYPE>(Node)) {
     Name = TYPE_NAME(Node);
   }
 
   if (Name) {
-    if (TYPE_STUB_DECL(Name)) {
+    if (isa<TYPE>(Name) && TYPE_STUB_DECL(Name)) {
       tree Stub = TYPE_STUB_DECL(Name);
       Location = expand_location(DECL_SOURCE_LOCATION(Stub));
     } else if (DECL_P(Name)) {
@@ -175,7 +178,7 @@ static expanded_location GetNodeLocation(tree Node, bool UseStub = true) {
   }
 
   if (!Location.line) {
-    if (UseStub && TYPE_STUB_DECL(Node)) {
+    if (UseStub && isa<TYPE>(Node) && TYPE_STUB_DECL(Node)) {
       tree Stub = TYPE_STUB_DECL(Node);
       Location = expand_location(DECL_SOURCE_LOCATION(Stub));
     } else if (DECL_P(Node)) {
@@ -209,11 +212,10 @@ DebugInfo::DebugInfo(Module *m)
 : M(m)
 , DebugFactory(*m)
 , CurFullPath("")
-, CurLineNo(0)
 , PrevFullPath("")
+, CurLineNo(0)
 , PrevLineNo(0)
 , PrevBB(NULL)
-, RegionStack()
 {}
 
 /// getFunctionName - Get function name for the given FnDecl. If the
@@ -289,7 +291,7 @@ void DebugInfo::EmitFunctionStart(tree FnDecl, Function *Fn) {
   unsigned VIndex = 0;
   DIType ContainingType;
   if (DECL_VINDEX (FnDecl) &&
-      DECL_CONTEXT (FnDecl) && TYPE_P((DECL_CONTEXT (FnDecl)))) { // Workaround GCC PR42653
+      DECL_CONTEXT (FnDecl) && isa<TYPE>((DECL_CONTEXT (FnDecl)))) { // Workaround GCC PR42653
     if (host_integerp (DECL_VINDEX (FnDecl), 0))
       VIndex = tree_low_cst (DECL_VINDEX (FnDecl), 0);
     Virtuality = dwarf::DW_VIRTUALITY_virtual;
@@ -342,11 +344,11 @@ DIDescriptor DebugInfo::findRegion(tree Node) {
     if (MDNode *R = dyn_cast_or_null<MDNode>(&*I->second))
       return DIDescriptor(R);
 
-  if (TYPE_P (Node)) {
+  if (isa<TYPE>(Node)) {
     DIType Ty = getOrCreateType(Node);
     return DIDescriptor(Ty);
   } else if (DECL_P (Node)) {
-    if (TREE_CODE (Node) == NAMESPACE_DECL) {
+    if (isa<NAMESPACE_DECL>(Node)) {
       DIDescriptor NSContext = findRegion(DECL_CONTEXT(Node));
       DINameSpace NS = getOrCreateNameSpace(Node, NSContext);
       return DIDescriptor(NS);
@@ -372,7 +374,7 @@ void DebugInfo::EmitFunctionEnd(bool EndFunction) {
 }
 
 /// EmitDeclare - Constructs the debug code for allocation of a new variable.
-void DebugInfo::EmitDeclare(tree decl, unsigned Tag, const char *Name,
+void DebugInfo::EmitDeclare(tree decl, unsigned Tag, StringRef Name,
                             tree type, Value *AI, LLVMBuilder &Builder) {
 
   // Ignore compiler generated temporaries.
@@ -386,7 +388,7 @@ void DebugInfo::EmitDeclare(tree decl, unsigned Tag, const char *Name,
   // Construct variable.
   DIScope VarScope = DIScope(cast<MDNode>(RegionStack.back()));
   DIType Ty = getOrCreateType(type);
-  if (!Ty && TREE_CODE(type) == OFFSET_TYPE)
+  if (!Ty && isa<OFFSET_TYPE>(type))
     Ty = createPointerType(TREE_TYPE(type));
   if (Ty && DECL_ARTIFICIAL (decl))
       Ty = DebugFactory.CreateArtificialType(Ty);
@@ -404,7 +406,7 @@ void DebugInfo::EmitDeclare(tree decl, unsigned Tag, const char *Name,
   Call->setDebugLoc(DebugLoc::get(Loc.line, 0, VarScope));
 }
 
-/// EmitStopPoint - Set current source location. 
+/// EmitStopPoint - Set current source location.
 void DebugInfo::EmitStopPoint(BasicBlock *CurBB, LLVMBuilder &Builder) {
   // Don't bother if things are the same as last time.
   if (PrevLineNo == CurLineNo &&
@@ -440,7 +442,7 @@ void DebugInfo::EmitGlobalVariable(GlobalVariable *GV, tree decl) {
   StringRef LinkageName;
   // The gdb does not expect linkage names for function local statics.
   if (DECL_CONTEXT (decl))
-    if (TREE_CODE (DECL_CONTEXT (decl)) != FUNCTION_DECL)
+    if (!isa<FUNCTION_DECL>(DECL_CONTEXT (decl)))
       LinkageName = GV->getName();
   DebugFactory.CreateGlobalVariable(findRegion(DECL_CONTEXT(decl)),
                                     DispName, DispName, LinkageName,
@@ -475,7 +477,7 @@ DIType DebugInfo::createBasicType(tree type) {
     Encoding = DW_ATE_float;
     break;
   case COMPLEX_TYPE:
-    Encoding = TREE_CODE(TREE_TYPE(type)) == REAL_TYPE ?
+    Encoding = isa<REAL_TYPE>(TREE_TYPE(type)) ?
       DW_ATE_complex_float : DW_ATE_lo_user;
     break;
   case BOOLEAN_TYPE:
@@ -501,14 +503,12 @@ DIType DebugInfo::createBasicType(tree type) {
 /// isArtificialArgumentType - Return true if arg_type represents artificial,
 /// i.e. "this" in c++, argument.
 static bool isArtificialArgumentType(tree arg_type, tree method_type) {
-  if (TREE_CODE (method_type) != METHOD_TYPE) return false;
-  if (TREE_CODE (arg_type) != POINTER_TYPE) return false;
+  if (!isa<METHOD_TYPE>(method_type)) return false;
+  if (!isa<POINTER_TYPE>(arg_type)) return false;
   if (TREE_TYPE (arg_type) == TYPE_METHOD_BASETYPE (method_type))
     return true;
-  if (TYPE_MAIN_VARIANT (TREE_TYPE (arg_type))
-      && TYPE_MAIN_VARIANT (TREE_TYPE (arg_type)) != TREE_TYPE (arg_type)
-      && (TYPE_MAIN_VARIANT (TREE_TYPE (arg_type))
-          == TYPE_METHOD_BASETYPE (method_type)))
+  if (main_type (arg_type) && main_type (arg_type) != TREE_TYPE (arg_type)
+      && (main_type (arg_type) == TYPE_METHOD_BASETYPE (method_type)))
     return true;
   return false;
 }
@@ -575,13 +575,13 @@ DIType DebugInfo::createPointerType(tree type) {
   DIType FromTy = getOrCreateType(TREE_TYPE(type));
   // type* and type&
   // FIXME: Should BLOCK_POINTER_TYP have its own DW_TAG?
-  unsigned Tag = TREE_CODE(type) == REFERENCE_TYPE ?
+  unsigned Tag = isa<REFERENCE_TYPE>(type) ?
     DW_TAG_reference_type: DW_TAG_pointer_type;
   unsigned Flags = 0;
 
   // Check if this pointer type has a name.
   if (tree TyName = TYPE_NAME(type))
-    if (TREE_CODE(TyName) == TYPE_DECL && !DECL_ORIGINAL_TYPE(TyName)) {
+    if (isa<TYPE_DECL>(TyName) && !DECL_ORIGINAL_TYPE(TyName)) {
       expanded_location TypeNameLoc = GetNodeLocation(TyName);
       DIType Ty =
         DebugFactory.CreateDerivedType(Tag, findRegion(DECL_CONTEXT(TyName)),
@@ -616,8 +616,8 @@ DIType DebugInfo::createPointerType(tree type) {
 DIType DebugInfo::createArrayType(tree type) {
 
   // type[n][m]...[p]
-  if (TREE_CODE (type) == ARRAY_TYPE
-      && TYPE_STRING_FLAG(type) && TREE_CODE(TREE_TYPE(type)) == INTEGER_TYPE){
+  if (isa<ARRAY_TYPE>(type)
+      && TYPE_STRING_FLAG(type) && isa<INTEGER_TYPE>(TREE_TYPE(type))) {
     DEBUGASSERT(0 && "Don't support pascal strings");
     return DIType();
   }
@@ -630,9 +630,9 @@ DIType DebugInfo::createArrayType(tree type) {
   // There will be ARRAY_TYPE nodes for each rank.  Followed by the derived
   // type.
   tree EltTy = TREE_TYPE(type);
-  if (TREE_CODE(type) == ARRAY_TYPE) {
+  if (isa<ARRAY_TYPE>(type)) {
     tree atype = type;
-    for (; TREE_CODE(atype) == ARRAY_TYPE; atype = TREE_TYPE(atype)) {
+    for (; isa<ARRAY_TYPE>(atype); atype = TREE_TYPE(atype)) {
       tree Domain = TYPE_DOMAIN(atype);
       if (Domain) {
         // FIXME - handle dynamic ranges
@@ -649,7 +649,7 @@ DIType DebugInfo::createArrayType(tree type) {
       EltTy = TREE_TYPE(atype);
     }
   } else {
-    assert(TREE_CODE(type) == VECTOR_TYPE && "Not an array or vector type!");
+    assert(isa<VECTOR_TYPE>(type) && "Not an array or vector type!");
     unsigned Length = TYPE_VECTOR_SUBPARTS(type);
     Subscripts.push_back(DebugFactory.GetOrCreateSubrange(0, Length));
   }
@@ -675,7 +675,7 @@ DIType DebugInfo::createEnumType(tree type) {
   if (TYPE_SIZE(type)) {
     for (tree Link = TYPE_VALUES(type); Link; Link = TREE_CHAIN(Link)) {
       tree EnumValue = TREE_VALUE(Link);
-      if (TREE_CODE(EnumValue) == CONST_DECL)
+      if (isa<CONST_DECL>(EnumValue))
         EnumValue = DECL_INITIAL(EnumValue);
       int64_t Value = getInt64(EnumValue, false);
       const char *EnumName = IDENTIFIER_POINTER(TREE_PURPOSE(Link));
@@ -705,7 +705,7 @@ DIType DebugInfo::createEnumType(tree type) {
 DIType DebugInfo::createStructType(tree type) {
 
   // struct { a; b; ... z; }; | union { a; b; ... z; };
-  unsigned Tag = TREE_CODE(type) == RECORD_TYPE ? DW_TAG_structure_type :
+  unsigned Tag = isa<RECORD_TYPE>(type) ? DW_TAG_structure_type :
     DW_TAG_union_type;
 
   unsigned RunTimeLang = 0;
@@ -741,10 +741,12 @@ DIType DebugInfo::createStructType(tree type) {
 
   // Check if this type is created while creating context information
   // descriptor.
-  std::map<tree_node *, WeakVH >::iterator I = TypeCache.find(type);
-  if (I != TypeCache.end())
-    if (MDNode *TN = dyn_cast_or_null<MDNode>(&*I->second))
-      return DIType(TN);
+  {
+    std::map<tree_node *, WeakVH >::iterator I = TypeCache.find(type);
+    if (I != TypeCache.end())
+      if (MDNode *TN = dyn_cast_or_null<MDNode>(&*I->second))
+        return DIType(TN);
+  }
 
   // forward declaration,
   if (TYPE_SIZE(type) == 0) {
@@ -820,7 +822,7 @@ DIType DebugInfo::createStructType(tree type) {
     // Get the location of the member.
     expanded_location MemLoc = GetNodeLocation(Member, false);
 
-    if (TREE_CODE(Member) != FIELD_DECL)
+    if (!isa<FIELD_DECL>(Member))
       // otherwise is a static variable, whose debug info is emitted
       // when through EmitGlobalVariable().
       continue;
@@ -831,9 +833,8 @@ DIType DebugInfo::createStructType(tree type) {
       continue;
 
     /* Ignore nameless fields.  */
-    if (DECL_NAME (Member) == NULL_TREE
-        && !(TREE_CODE (TREE_TYPE (Member)) == UNION_TYPE
-             || TREE_CODE (TREE_TYPE (Member)) == RECORD_TYPE))
+    if (DECL_NAME (Member) == NULL_TREE &&
+        !isa<RECORD_OR_UNION_TYPE>(TREE_TYPE(Member)))
       continue;
 
     // Field type is the declared type of the field.
@@ -933,11 +934,11 @@ DIType DebugInfo::createVariantType(tree type, DIType MainTy) {
 
   DIType Ty;
   if (tree TyDef = TYPE_NAME(type)) {
-      std::map<tree_node *, WeakVH >::iterator I = TypeCache.find(TyDef);
-      if (I != TypeCache.end())
-        if (Value *M = I->second)
-          return DIType(cast<MDNode>(M));
-    if (TREE_CODE(TyDef) == TYPE_DECL &&  DECL_ORIGINAL_TYPE(TyDef)) {
+    std::map<tree_node *, WeakVH >::iterator I = TypeCache.find(TyDef);
+    if (I != TypeCache.end())
+      if (I->second)
+        return DIType(cast<MDNode>(I->second));
+    if (isa<TYPE_DECL>(TyDef) &&  DECL_ORIGINAL_TYPE(TyDef)) {
       expanded_location TypeDefLoc = GetNodeLocation(TyDef);
       Ty = DebugFactory.CreateDerivedType(DW_TAG_typedef,
                                           findRegion(DECL_CONTEXT(TyDef)),
@@ -998,13 +999,13 @@ DIType DebugInfo::getOrCreateType(tree type) {
 
   // Should only be void if a pointer/reference/return type.  Returning NULL
   // allows the caller to produce a non-derived type.
-  if (TREE_CODE(type) == VOID_TYPE) return DIType();
+  if (isa<VOID_TYPE>(type)) return DIType();
 
   // Check to see if the compile unit already has created this type.
   std::map<tree_node *, WeakVH >::iterator I = TypeCache.find(type);
   if (I != TypeCache.end())
-    if (Value *M = I->second)
-      return DIType(cast<MDNode>(M));
+    if (I->second)
+      return DIType(cast<MDNode>(I->second));
 
   DIType MainTy;
   if (type != TYPE_MAIN_VARIANT(type) && TYPE_MAIN_VARIANT(type))
@@ -1029,7 +1030,6 @@ DIType DebugInfo::getOrCreateType(tree type) {
       // Do not cache pointer type. The pointer may point to forward declared
       // struct.
       return createPointerType(type);
-      break;
 
     case OFFSET_TYPE: {
       // gen_type_die(TYPE_OFFSET_BASETYPE(type), context_die);
@@ -1057,7 +1057,6 @@ DIType DebugInfo::getOrCreateType(tree type) {
     case QUAL_UNION_TYPE:
     case UNION_TYPE:
       return createStructType(type);
-      break;
 
     case INTEGER_TYPE:
     case REAL_TYPE:
@@ -1134,8 +1133,7 @@ DICompileUnit DebugInfo::getOrCreateCompileUnit(const char *FullPath,
   unsigned ObjcRunTimeVer = 0;
 //  if (flag_objc_abi != 0 && flag_objc_abi != -1)
 //    ObjcRunTimeVer = flag_objc_abi;
-  return DebugFactory.CreateCompileUnit(LangTag, FileName.c_str(),
-                                        Directory.c_str(),
+  return DebugFactory.CreateCompileUnit(LangTag, FileName, Directory,
                                         version_string, isMain,
                                         optimize, Flags,
                                         ObjcRunTimeVer);
@@ -1169,7 +1167,7 @@ Constant *DIFactory::GetTagConstant(unsigned TAG) {
          "Tag too large for debug encoding!");
   // llvm has moved forward. DIFactory does not emit debug info in updated form.
   // Use LLVMDebugVersion10 directly here.
-  return ConstantInt::get(Type::getInt32Ty(VMContext), 
+  return ConstantInt::get(Type::getInt32Ty(VMContext),
                           TAG | LLVMDebugVersion10);
 }
 
@@ -1753,7 +1751,7 @@ Instruction *DIFactory::InsertDeclare(Value *Storage, DIVariable D,
   if (!DeclareFn)
     DeclareFn = Intrinsic::getDeclaration(&M, Intrinsic::dbg_declare);
 
-  Value *Args[] = { MDNode::get(Storage->getContext(), Storage), 
+  Value *Args[] = { MDNode::get(Storage->getContext(), Storage),
                     D };
   return CallInst::Create(DeclareFn, Args, "", InsertBefore);
 }
