@@ -1,6 +1,6 @@
 //===------------ DefaultABI.cpp - Default ABI implementation -------------===//
 //
-// Copyright (C) 2010, 2011  Rafael Espindola, Duncan Sands et al.
+// Copyright (C) 2010 to 2012  Rafael Espindola, Duncan Sands et al.
 //
 // This file is part of DragonEgg.
 //
@@ -27,22 +27,33 @@
 #include <gmp.h>
 
 // GCC headers
+#include "auto-host.h"
+#ifndef ENABLE_BUILD_WITH_CXX
 extern "C" {
+#endif
 #include "config.h"
 // Stop GCC declaring 'getopt' as it can clash with the system's declaration.
 #undef HAVE_DECL_GETOPT
 #include "system.h"
 #include "coretypes.h"
-#include "target.h"
+#include "tm.h"
 #include "tree.h"
-}
+#ifndef ENABLE_BUILD_WITH_CXX
+} // extern "C"
+#endif
+
+// Trees header.
+#include "dragonegg/Trees.h"
+
+void DefaultABIClient::anchor() {}
 
 // doNotUseShadowReturn - Return true if the specified GCC type
 // should not be returned using a pointer to struct parameter.
-bool doNotUseShadowReturn(tree type, tree fndecl, CallingConv::ID /*CC*/) {
+bool doNotUseShadowReturn(tree type, tree fndecl, CallingConv::ID CC) {
+  (void)CC; // Not used by all ABI macros.
   if (!TYPE_SIZE(type))
     return false;
-  if (TREE_CODE(TYPE_SIZE(type)) != INTEGER_CST)
+  if (!isa<INTEGER_CST>(TYPE_SIZE(type)))
     return false;
   // LLVM says do not use shadow argument.
   if (LLVM_SHOULD_NOT_RETURN_COMPLEX_IN_MEMORY(type) ||
@@ -63,9 +74,9 @@ bool doNotUseShadowReturn(tree type, tree fndecl, CallingConv::ID /*CC*/) {
 tree isSingleElementStructOrArray(tree type, bool ignoreZeroLength,
                                   bool rejectFatBitfield) {
   // Complex numbers have two fields.
-  if (TREE_CODE(type) == COMPLEX_TYPE) return 0;
+  if (isa<COMPLEX_TYPE>(type)) return 0;
   // All other scalars are good.
-  if (!AGGREGATE_TYPE_P(type)) return type;
+  if (!isa<AGGREGATE_TYPE>(type)) return type;
 
   tree FoundField = 0;
   switch (TREE_CODE(type)) {
@@ -76,20 +87,19 @@ tree isSingleElementStructOrArray(tree type, bool ignoreZeroLength,
     return 0;
   case RECORD_TYPE:
     // If this record has variable length, reject it.
-    if (TREE_CODE(TYPE_SIZE(type)) != INTEGER_CST)
+    if (!isa<INTEGER_CST>(TYPE_SIZE(type)))
       return 0;
 
     for (tree Field = TYPE_FIELDS(type); Field; Field = TREE_CHAIN(Field))
-      if (TREE_CODE(Field) == FIELD_DECL) {
+      if (isa<FIELD_DECL>(Field)) {
         if (ignoreZeroLength) {
-          if (DECL_SIZE(Field) &&
-              TREE_CODE(DECL_SIZE(Field)) == INTEGER_CST &&
+          if (DECL_SIZE(Field) && isa<INTEGER_CST>(DECL_SIZE(Field)) &&
               TREE_INT_CST_LOW(DECL_SIZE(Field)) == 0)
             continue;
         }
         if (!FoundField) {
           if (rejectFatBitfield &&
-              TREE_CODE(TYPE_SIZE(type)) == INTEGER_CST &&
+              isa<INTEGER_CST>(TYPE_SIZE(type)) &&
               TREE_INT_CST_LOW(TYPE_SIZE(TREE_TYPE(Field))) >
               TREE_INT_CST_LOW(TYPE_SIZE(type)))
             return 0;
@@ -112,9 +122,7 @@ tree isSingleElementStructOrArray(tree type, bool ignoreZeroLength,
 /// isZeroSizedStructOrUnion - Returns true if this is a struct or union
 /// which is zero bits wide.
 bool isZeroSizedStructOrUnion(tree type) {
-  if (TREE_CODE(type) != RECORD_TYPE &&
-      TREE_CODE(type) != UNION_TYPE &&
-      TREE_CODE(type) != QUAL_UNION_TYPE)
+  if (!isa<RECORD_OR_UNION_TYPE>(type))
     return false;
   return int_size_in_bytes(type) == 0;
 }
@@ -128,6 +136,7 @@ bool DefaultABI::isShadowReturn() const { return C.isShadowReturn(); }
 /// on the client that indicate how its pieces should be handled.  This
 /// handles things like returning structures via hidden parameters.
 void DefaultABI::HandleReturnType(tree type, tree fn, bool isBuiltin) {
+  (void)isBuiltin; // Not used by all ABI macros.
   unsigned Offset = 0;
   Type *Ty = ConvertType(type);
   if (Ty->isVectorTy()) {
@@ -147,7 +156,7 @@ void DefaultABI::HandleReturnType(tree type, tree fn, bool isBuiltin) {
   } else if (doNotUseShadowReturn(type, fn, C.getCallingConv())) {
     tree SingleElt = LLVM_SHOULD_RETURN_SELT_STRUCT_AS_SCALAR(type);
     if (SingleElt && TYPE_SIZE(SingleElt) &&
-        TREE_CODE(TYPE_SIZE(SingleElt)) == INTEGER_CST &&
+        isa<INTEGER_CST>(TYPE_SIZE(SingleElt)) &&
         TREE_INT_CST_LOW(TYPE_SIZE_UNIT(type)) ==
         TREE_INT_CST_LOW(TYPE_SIZE_UNIT(SingleElt))) {
       C.HandleAggregateResultAsScalar(ConvertType(SingleElt));
@@ -161,7 +170,7 @@ void DefaultABI::HandleReturnType(tree type, tree fn, bool isBuiltin) {
                LLVM_SCALAR_TYPE_FOR_STRUCT_RETURN(type, &Offset))
         C.HandleAggregateResultAsScalar(ScalarTy, Offset);
       else
-        DieAbjectly("Unable to determine how to return this aggregate!");
+        llvm_unreachable("Unable to determine how to return this aggregate!");
     }
   } else {
     // If the function is returning a struct or union, we pass the pointer to
@@ -169,7 +178,7 @@ void DefaultABI::HandleReturnType(tree type, tree fn, bool isBuiltin) {
 
     // FIXME: should return the hidden first argument for some targets
     // (e.g. ELF i386).
-    if (AGGREGATE_TYPE_P(type))
+    if (isa<AGGREGATE_TYPE>(type))
       C.HandleAggregateShadowResult(Ty->getPointerTo(), false);
     else
       C.HandleScalarShadowResult(Ty->getPointerTo(), false);
@@ -248,11 +257,10 @@ void DefaultABI::HandleArgument(tree type, std::vector<Type*> &ScalarElts,
   } else if (isZeroSizedStructOrUnion(type)) {
     // Zero sized struct or union, just drop it!
     ;
-  } else if (TREE_CODE(type) == RECORD_TYPE) {
+  } else if (isa<RECORD_TYPE>(type)) {
     for (tree Field = TYPE_FIELDS(type); Field; Field = TREE_CHAIN(Field))
-      if (TREE_CODE(Field) == FIELD_DECL) {
+      if (isa<FIELD_DECL>(Field)) {
         const tree Ftype = TREE_TYPE(Field);
-        Type *FTy = ConvertType(Ftype);
         unsigned FNo = GetFieldIndex(Field, Ty);
         assert(FNo < INT_MAX && "Case not handled yet!");
 
@@ -262,23 +270,25 @@ void DefaultABI::HandleArgument(tree type, std::vector<Type*> &ScalarElts,
         // (We know there currently are no other such cases active because
         // they would hit the assert in FunctionPrologArgumentConversion::
         // HandleByValArgument.)
+        Type *FTy = ConvertType(Ftype);
+        (void)FTy; // Not used by all ABI macros.
         if (!LLVM_SHOULD_PASS_AGGREGATE_USING_BYVAL_ATTR(Ftype, FTy)) {
           C.EnterField(FNo, Ty);
           HandleArgument(TREE_TYPE(Field), ScalarElts);
           C.ExitField();
         }
       }
-  } else if (TREE_CODE(type) == COMPLEX_TYPE) {
+  } else if (isa<COMPLEX_TYPE>(type)) {
     C.EnterField(0, Ty);
     HandleArgument(TREE_TYPE(type), ScalarElts);
     C.ExitField();
     C.EnterField(1, Ty);
     HandleArgument(TREE_TYPE(type), ScalarElts);
     C.ExitField();
-  } else if ((TREE_CODE(type) == UNION_TYPE) ||
-             (TREE_CODE(type) == QUAL_UNION_TYPE)) {
+  } else if ((isa<UNION_TYPE>(type)) ||
+             (isa<QUAL_UNION_TYPE>(type))) {
     HandleUnion(type, ScalarElts);
-  } else if (TREE_CODE(type) == ARRAY_TYPE) {
+  } else if (isa<ARRAY_TYPE>(type)) {
     // Array with padding?
     if (Ty->isStructTy())
       Ty = cast<StructType>(Ty)->getTypeAtIndex(0U);
@@ -289,7 +299,7 @@ void DefaultABI::HandleArgument(tree type, std::vector<Type*> &ScalarElts,
       C.ExitField();
     }
   } else {
-    DieAbjectly("Unknown aggregate type!");
+    llvm_unreachable("Unknown aggregate type!");
   }
 }
 
@@ -298,7 +308,7 @@ void DefaultABI::HandleUnion(tree type, std::vector<Type*> &ScalarElts) {
   if (TYPE_TRANSPARENT_AGGR(type)) {
     tree Field = TYPE_FIELDS(type);
     assert(Field && "Transparent union must have some elements!");
-    while (TREE_CODE(Field) != FIELD_DECL) {
+    while (!isa<FIELD_DECL>(Field)) {
       Field = TREE_CHAIN(Field);
       assert(Field && "Transparent union must have some elements!");
     }
@@ -309,23 +319,13 @@ void DefaultABI::HandleUnion(tree type, std::vector<Type*> &ScalarElts) {
     unsigned MaxSize = 0;
     tree MaxElt = 0;
     for (tree Field = TYPE_FIELDS(type); Field; Field = TREE_CHAIN(Field)) {
-      if (TREE_CODE(Field) == FIELD_DECL) {
-        // Skip fields that are known not to be present.
-        if (TREE_CODE(type) == QUAL_UNION_TYPE &&
-            integer_zerop(DECL_QUALIFIER(Field)))
-          continue;
-
+      if (isa<FIELD_DECL>(Field)) {
         tree SizeTree = TYPE_SIZE(TREE_TYPE(Field));
         unsigned Size = ((unsigned)TREE_INT_CST_LOW(SizeTree)+7)/8;
         if (Size > MaxSize) {
           MaxSize = Size;
           MaxElt = Field;
         }
-
-        // Skip remaining fields if this one is known to be present.
-        if (TREE_CODE(type) == QUAL_UNION_TYPE &&
-            integer_onep(DECL_QUALIFIER(Field)))
-          break;
       }
     }
 
