@@ -1,7 +1,6 @@
 //===----------- Types.cpp - Converting GCC types to LLVM types -----------===//
 //
-// Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011  Chris Lattner,
-// Duncan Sands et al.
+// Copyright (C) 2005 to 2012  Chris Lattner, Duncan Sands et al.
 //
 // This file is part of DragonEgg.
 //
@@ -29,29 +28,29 @@
 #include "dragonegg/ADT/Range.h"
 
 // LLVM headers
-#include "llvm/Module.h"
-#include "llvm/ADT/GraphTraits.h"
 #include "llvm/ADT/SCCIterator.h"
-#include "llvm/Assembly/Writer.h"
-#include "llvm/Support/ErrorHandling.h"
-#include "llvm/Support/raw_ostream.h"
-#include "llvm/Target/TargetMachine.h"
 
 // System headers
 #include <gmp.h>
 #include <map>
 
 // GCC headers
+#include "auto-host.h"
+#ifndef ENABLE_BUILD_WITH_CXX
 extern "C" {
+#endif
 #include "config.h"
 // Stop GCC declaring 'getopt' as it can clash with the system's declaration.
 #undef HAVE_DECL_GETOPT
 #include "system.h"
 #include "coretypes.h"
-#include "target.h"
+#include "tm.h"
 #include "tree.h"
+
 #include "flags.h"
-}
+#ifndef ENABLE_BUILD_WITH_CXX
+} // extern "C"
+#endif
 
 // Trees header.
 #include "dragonegg/Trees.h"
@@ -85,7 +84,7 @@ namespace {
   public:
     /// Dereference operator.
     tree operator*() {
-      return TREE_CODE(type_ref) == TREE_LIST ?
+      return isa<TREE_LIST>(type_ref) ?
         TREE_VALUE(type_ref) : TREE_TYPE(type_ref);
     };
 
@@ -97,20 +96,14 @@ namespace {
       return !(*this == other);
     }
 
-    /// Postfix increment operator.
-    ContainedTypeIterator operator++(int) {
-      ContainedTypeIterator Result(*this);
-      ++(*this);
-      return Result;
-    }
-
     /// Prefix increment operator.
     ContainedTypeIterator& operator++() {
       assert(type_ref && "Incrementing end iterator!");
 
       switch (TREE_CODE(type_ref)) {
       default:
-        assert(false && "Unexpected tree kind!");
+        debug_tree(type_ref);
+        llvm_unreachable("Unexpected tree kind!");
       case ARRAY_TYPE:
       case COMPLEX_TYPE:
       case POINTER_TYPE:
@@ -124,7 +117,9 @@ namespace {
       case FIELD_DECL:
         // Here type_ref is a field of the record or union type being iterated
         // over.  Move on to the next field.
-        type_ref = TREE_CHAIN(type_ref);
+        do
+          type_ref = TREE_CHAIN(type_ref);
+        while (type_ref && !isa<FIELD_DECL>(type_ref));
         break;
 
       case FUNCTION_TYPE:
@@ -154,7 +149,7 @@ namespace {
     static ContainedTypeIterator begin(tree type) {
       switch (TREE_CODE(type)) {
       default:
-        assert(false && "Unknown type!");
+        llvm_unreachable("Unknown type!");
 
       case BOOLEAN_TYPE:
       case ENUMERAL_TYPE:
@@ -179,7 +174,10 @@ namespace {
       case UNION_TYPE:
         // The contained types are the types of the record's fields.  Use the
         // first FIELD_DECL as the "pointer" to the first contained type.
-        return ContainedTypeIterator(TYPE_FIELDS(type));
+        for (tree field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field))
+          if (isa<FIELD_DECL>(field))
+            return ContainedTypeIterator(field);
+        return end();
 
       case FUNCTION_TYPE:
       case METHOD_TYPE:
@@ -210,7 +208,7 @@ namespace {
 /// ArrayLengthOf - Returns the length of the given gcc array type, or NO_LENGTH
 /// if the array has variable or unknown length.
 uint64_t ArrayLengthOf(tree type) {
-  assert(TREE_CODE(type) == ARRAY_TYPE && "Only for array types!");
+  assert(isa<ARRAY_TYPE>(type) && "Only for array types!");
   // Workaround for missing sanity checks in older versions of GCC.
   if ((GCC_MINOR == 5 && GCC_MICRO < 3) || (GCC_MINOR == 6 && GCC_MICRO < 2))
     if (!TYPE_DOMAIN(type) || !TYPE_MAX_VALUE(TYPE_DOMAIN(type)))
@@ -219,7 +217,7 @@ uint64_t ArrayLengthOf(tree type) {
   // Bail out if the array has variable or unknown length.
   if (!isInt64(range, false))
     return NO_LENGTH;
-  int64_t Range = getInt64(range, false);
+  int64_t Range = (int64_t)getInt64(range, false);
   return Range < 0 ? 0 : 1 + (uint64_t)Range;
 }
 
@@ -246,7 +244,7 @@ static int get_decl_index(tree t) {
 /// that its first bit is within the byte the LLVM field starts at).  Returns
 /// INT_MAX if there is no such LLVM field.
 int GetFieldIndex(tree decl, Type *Ty) {
-  assert(TREE_CODE(decl) == FIELD_DECL && "Expected a FIELD_DECL!");
+  assert(isa<FIELD_DECL>(decl) && "Expected a FIELD_DECL!");
   // FIXME: The following test sometimes fails when compiling Fortran90 because
   // DECL_CONTEXT does not point to the containing type, but some other type!
 //  assert(Ty == ConvertType(DECL_CONTEXT(decl)) && "Field not for this type!");
@@ -297,7 +295,7 @@ int GetFieldIndex(tree decl, Type *Ty) {
 /// getPointerToType - Returns the LLVM register type to use for a pointer to
 /// the given GCC type.
 Type *getPointerToType(tree type) {
-  if (VOID_TYPE_P(type))
+  if (isa<VOID_TYPE>(type))
     // void* -> byte*
     return GetUnitPointerType(Context);
   // FIXME: Handle address spaces.
@@ -329,7 +327,7 @@ Type *GetUnitPointerType(LLVMContext &C, unsigned AddrSpace) {
 /// that this returns false for function types, for which the GCC type size
 /// doesn't represent anything useful for us.
 static bool isSized(tree type) {
-  if (TREE_CODE(type) == FUNCTION_TYPE || TREE_CODE(type) == METHOD_TYPE)
+  if (isa<FUNCTION_TYPE>(type) || isa<METHOD_TYPE>(type))
     return false;
   return TYPE_SIZE(type);
 }
@@ -384,7 +382,7 @@ static Type *CheckTypeConversion(tree type, Type *Ty) {
     debug_tree(type);
     errs() << "LLVM: ";
     Ty->print(errs());
-    DieAbjectly("\nLLVM type doesn't represent GCC type!");
+    llvm_unreachable("\nLLVM type doesn't represent GCC type!");
   }
 #endif
 
@@ -416,7 +414,7 @@ bool isPassedByInvisibleReference(tree Type) {
   // FIXME: Search for TREE_ADDRESSABLE in calls.c, and see if there are other
   // cases that make arguments automatically passed in by reference.
   return TREE_ADDRESSABLE(Type) || TYPE_SIZE(Type) == 0 ||
-         TREE_CODE(TYPE_SIZE(Type)) != INTEGER_CST;
+         !isa<INTEGER_CST>(TYPE_SIZE(Type));
 }
 
 
@@ -428,15 +426,24 @@ bool isPassedByInvisibleReference(tree Type) {
 /// of the scalar GCC type 'type'.  All of the EmitReg* routines use this to
 /// determine the LLVM type to return.
 Type *getRegType(tree type) {
+  // Check that the type mode doesn't depend on the type variant (various bits
+  // of the plugin rely on this).
+  assert(TYPE_MODE(type) == TYPE_MODE(TYPE_MAIN_VARIANT(type))
+         && "Type mode differs between variants!");
+
+  // LLVM doesn't care about variants such as const, volatile, or restrict.
+  type = TYPE_MAIN_VARIANT(type);
+
   // NOTE: Any changes made here need to be reflected in LoadRegisterFromMemory,
   // StoreRegisterToMemory and ExtractRegisterFromConstant.
-  assert(!AGGREGATE_TYPE_P(type) && "Registers must have a scalar type!");
-  assert(TREE_CODE(type) != VOID_TYPE && "Registers cannot have void type!");
+  assert(!isa<AGGREGATE_TYPE>(type) && "Registers must have a scalar type!");
+  assert(!isa<VOID_TYPE>(type) && "Registers cannot have void type!");
 
   switch (TREE_CODE(type)) {
 
   default:
-    DieAbjectly("Unknown register type!", type);
+    debug_tree(type);
+    llvm_unreachable("Unknown register type!");
 
   case BOOLEAN_TYPE:
   case ENUMERAL_TYPE:
@@ -457,7 +464,7 @@ Type *getRegType(tree type) {
   case POINTER_TYPE:
   case REFERENCE_TYPE:
     // void* -> byte*
-    return VOID_TYPE_P(TREE_TYPE(type)) ?  GetUnitPointerType(Context) :
+    return isa<VOID_TYPE>(TREE_TYPE(type)) ?  GetUnitPointerType(Context) :
       ConvertType(TREE_TYPE(type))->getPointerTo();
 
   case REAL_TYPE:
@@ -474,12 +481,13 @@ Type *getRegType(tree type) {
       // IEEE quad precision.
       return Type::getFP128Ty(Context);
 #endif
-      DieAbjectly("Unknown FP type!", type);
+    debug_tree(type);
+    llvm_unreachable("Unknown FP type!");
 
   case VECTOR_TYPE: {
     // LLVM does not support vectors of pointers, so turn any pointers into
     // integers.
-    Type *EltTy = POINTER_TYPE_P(TREE_TYPE(type)) ?
+    Type *EltTy = isa<ACCESS_TYPE>(TREE_TYPE(type)) ?
       getTargetData().getIntPtrType(Context) : getRegType(TREE_TYPE(type));
     return VectorType::get(EltTy, TYPE_VECTOR_SUBPARTS(type));
   }
@@ -503,7 +511,7 @@ static Type *ConvertArrayTypeRecursive(tree type) {
   Type *Ty = ArrayType::get(ElementTy, NumElements);
 
   // If the array is underaligned, wrap it in a packed struct.
-  if (TYPE_ALIGN(type) < getTargetData().getABITypeAlignment(Ty) * 8)
+  if (TYPE_ALIGN(type) < TYPE_ALIGN(main_type(type)))
     Ty = StructType::get(Context, Ty, /*isPacked*/ true);
 
   // If the user increased the alignment of the array element type, then the
@@ -529,14 +537,14 @@ namespace {
     Type *&RetTy;
     SmallVectorImpl<Type*> &ArgTypes;
     CallingConv::ID &CallingConv;
+    unsigned Offset;
     bool isShadowRet;
     bool KNRPromotion;
-    unsigned Offset;
   public:
     FunctionTypeConversion(Type *&retty, SmallVectorImpl<Type*> &AT,
                            CallingConv::ID &CC, bool KNR)
-      : RetTy(retty), ArgTypes(AT), CallingConv(CC), KNRPromotion(KNR),
-        Offset(0) {
+      : RetTy(retty), ArgTypes(AT), CallingConv(CC), Offset(0),
+        KNRPromotion(KNR) {
       CallingConv = CallingConv::C;
       isShadowRet = false;
     }
@@ -548,16 +556,16 @@ namespace {
 
     /// HandleScalarResult - This callback is invoked if the function returns a
     /// simple scalar result value.
-    void HandleScalarResult(Type *RetTy) {
-      this->RetTy = RetTy;
+    void HandleScalarResult(Type *RTy) {
+      this->RetTy = RTy;
     }
 
     /// HandleAggregateResultAsScalar - This callback is invoked if the function
     /// returns an aggregate value by bit converting it to the specified scalar
     /// type and returning that.
-    void HandleAggregateResultAsScalar(Type *ScalarTy, unsigned Offset=0) {
+    void HandleAggregateResultAsScalar(Type *ScalarTy, unsigned Off=0) {
       RetTy = ScalarTy;
-      this->Offset = Offset;
+      this->Offset = Off;
     }
 
     /// HandleAggregateResultAsAggregate - This callback is invoked if the function
@@ -635,10 +643,10 @@ namespace {
 }
 
 static Attributes HandleArgumentExtension(tree ArgTy) {
-  if (TREE_CODE(ArgTy) == BOOLEAN_TYPE) {
+  if (isa<BOOLEAN_TYPE>(ArgTy)) {
     if (TREE_INT_CST_LOW(TYPE_SIZE(ArgTy)) < INT_TYPE_SIZE)
       return Attribute::ZExt;
-  } else if (TREE_CODE(ArgTy) == INTEGER_TYPE &&
+  } else if (isa<INTEGER_TYPE>(ArgTy) &&
              TREE_INT_CST_LOW(TYPE_SIZE(ArgTy)) < INT_TYPE_SIZE) {
     if (TYPE_UNSIGNED(ArgTy))
       return Attribute::ZExt;
@@ -712,6 +720,10 @@ FunctionType *ConvertArgListToFnType(tree type, ArrayRef<tree> Args,
     // Compute zext/sext attributes.
     PAttributes |= HandleArgumentExtension(ArgTy);
 
+    // Compute noalias attributes.
+    if (isa<ACCESS_TYPE>(ArgTy) && TYPE_RESTRICT(ArgTy))
+      PAttributes |= Attribute::NoAlias;
+
     if (PAttributes != Attribute::None)
       Attrs.push_back(AttributeWithIndex::get(ArgTys.size(), PAttributes));
   }
@@ -805,11 +817,11 @@ FunctionType *ConvertFunctionType(tree type, tree decl, tree static_chain,
                                              Attribute::Nest));
   }
 
+#ifdef LLVM_TARGET_ENABLE_REGPARM
   // If the target has regparam parameters, allow it to inspect the function
   // type.
   int local_regparam = 0;
   int local_fp_regparam = 0;
-#ifdef LLVM_TARGET_ENABLE_REGPARM
   LLVM_TARGET_INIT_REGPARM(local_regparam, local_fp_regparam, type);
 #endif // LLVM_TARGET_ENABLE_REGPARM
 
@@ -851,23 +863,20 @@ FunctionType *ConvertFunctionType(tree type, tree decl, tree static_chain,
     // inspect it for restrict qualifiers, otherwise try the argument
     // types.
     tree RestrictArgTy = (DeclArgs) ? TREE_TYPE(DeclArgs) : ArgTy;
-    if (TREE_CODE(RestrictArgTy) == POINTER_TYPE ||
-        TREE_CODE(RestrictArgTy) == REFERENCE_TYPE) {
-      if (TYPE_RESTRICT(RestrictArgTy))
-        PAttributes |= Attribute::NoAlias;
-    }
+    if (isa<ACCESS_TYPE>(RestrictArgTy) && TYPE_RESTRICT(RestrictArgTy))
+      PAttributes |= Attribute::NoAlias;
 
 #ifdef LLVM_TARGET_ENABLE_REGPARM
     // Allow the target to mark this as inreg.
-    if (INTEGRAL_TYPE_P(ArgTy) || POINTER_TYPE_P(ArgTy) ||
-        SCALAR_FLOAT_TYPE_P(ArgTy))
+    if (isa<INTEGRAL_TYPE>(ArgTy) || isa<ACCESS_TYPE>(ArgTy) ||
+        isa<REAL_TYPE>(ArgTy))
       LLVM_ADJUST_REGPARM_ATTRIBUTE(PAttributes, ArgTy,
                                     TREE_INT_CST_LOW(TYPE_SIZE(ArgTy)),
                                     local_regparam, local_fp_regparam);
 #endif // LLVM_TARGET_ENABLE_REGPARM
 
     if (PAttributes != Attribute::None) {
-      HasByVal |= PAttributes & Attribute::ByVal;
+      HasByVal |= (PAttributes & Attribute::ByVal) != Attribute::None;
 
       // If the argument is split into multiple scalars, assign the
       // attributes to all scalars of the aggregate.
@@ -902,7 +911,7 @@ static Type *ConvertPointerTypeRecursive(tree type) {
   // This is where self-recursion loops are broken, by not converting the type
   // pointed to if this would cause trouble (the pointer type is turned into
   // {}* instead).
-  tree pointee = TYPE_MAIN_VARIANT(TREE_TYPE(type));
+  tree pointee = main_type(type);
 
   // The pointer type is in the strongly connected component (SCC) currently
   // being converted.  Check whether the pointee is as well.  If there is more
@@ -936,14 +945,12 @@ static Type *ConvertPointerTypeRecursive(tree type) {
 
     // Drill down through nested arrays to the ultimate element type.  Thanks
     // to this we may return S* for a (S[])*, which is better than {}*.
-    while (TREE_CODE(pointee) == ARRAY_TYPE)
-      pointee = TYPE_MAIN_VARIANT(TREE_TYPE(pointee));
+    while (isa<ARRAY_TYPE>(pointee))
+      pointee = main_type(pointee);
 
     // If the pointee is a record or union type then return a pointer to its
     // placeholder type.  Otherwise return {}*.
-    if (TREE_CODE(pointee) == QUAL_UNION_TYPE ||
-        TREE_CODE(pointee) == RECORD_TYPE ||
-        TREE_CODE(pointee) == UNION_TYPE)
+    if (isa<RECORD_OR_UNION_TYPE>(pointee))
       PointeeTy = getCachedType(pointee);
     else
       PointeeTy = StructType::get(Context);
@@ -1091,15 +1098,11 @@ static Type *ConvertRecordTypeRecursive(tree type) {
   // Record all interesting fields so they can easily be visited backwards.
   SmallVector<tree, 16> Fields;
   for (tree field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
-    assert(TREE_CODE(field) == FIELD_DECL && "Lang data not freed?");
+    if (!isa<FIELD_DECL>(field)) continue;
     // Ignore fields with variable or unknown position since they cannot be
     // represented by the LLVM type system.
     if (!OffsetIsLLVMCompatible(field))
       continue;
-    // Skip fields that are known not to be present.
-    if (TREE_CODE(type) == QUAL_UNION_TYPE &&
-        integer_zerop(DECL_QUALIFIER(field)))
-      continue;
     Fields.push_back(field);
   }
 
@@ -1232,7 +1235,7 @@ static bool mayRecurse(tree type) {
   assert(type == TYPE_MAIN_VARIANT(type) && "Not converting the main variant!");
   switch (TREE_CODE(type)) {
   default:
-    assert(false && "Unknown type!");
+    llvm_unreachable("Unknown type!");
 
   case BOOLEAN_TYPE:
   case ENUMERAL_TYPE:
@@ -1304,18 +1307,21 @@ static Type *ConvertTypeRecursive(tree type) {
   // (SCC) of the type graph.  This should always be the case because SCCs are
   // visited bottom up.
   bool inSCC = false;
-  for (unsigned i = 0, e = SCCInProgress->size(); i != e; ++i)
+  for (size_t i = 0, e = SCCInProgress->size(); i != e; ++i)
     if ((*SCCInProgress)[i] == type) {
       inSCC = true;
       break;
     }
-  if (!inSCC)
-    DieAbjectly("Type not in SCC!", type);
+  if (!inSCC) {
+    debug_tree(type);
+    llvm_unreachable("Type not in SCC!");
+  }
 #endif
 
   switch (TREE_CODE(type)) {
   default:
-    DieAbjectly("Unexpected type!", type);
+    debug_tree(type);
+    llvm_unreachable("Unexpected type!");
 
   case ARRAY_TYPE:
     return RememberTypeConversion(type, ConvertArrayTypeRecursive(type));
@@ -1348,7 +1354,8 @@ static Type *ConvertTypeNonRecursive(tree type) {
 
   switch (TREE_CODE(type)) {
   default:
-    DieAbjectly("Unknown or recursive type!", type);
+    debug_tree(type);
+    llvm_unreachable("Unknown or recursive type!");
 
   case ARRAY_TYPE:
   case FUNCTION_TYPE:
@@ -1376,7 +1383,7 @@ static Type *ConvertTypeNonRecursive(tree type) {
 
   case COMPLEX_TYPE: {
     if (Type *Ty = getCachedType(type)) return Ty;
-    Type *Ty = ConvertTypeNonRecursive(TYPE_MAIN_VARIANT(TREE_TYPE(type)));
+    Type *Ty = ConvertTypeNonRecursive(main_type(type));
     Ty = StructType::get(Ty, Ty, NULL);
     return RememberTypeConversion(type, Ty);
   }
@@ -1392,7 +1399,8 @@ static Type *ConvertTypeNonRecursive(tree type) {
     // Caching the type conversion is not worth it.
     switch (TYPE_PRECISION(type)) {
     default:
-      DieAbjectly("Unknown FP type!", type);
+      debug_tree(type);
+      llvm_unreachable("Unknown FP type!");
     case 32: return CheckTypeConversion(type, Type::getFloatTy(Context));
     case 64: return CheckTypeConversion(type, Type::getDoubleTy(Context));
     case 80: return CheckTypeConversion(type, Type::getX86_FP80Ty(Context));
@@ -1423,10 +1431,10 @@ static Type *ConvertTypeNonRecursive(tree type) {
     Type *Ty;
     // LLVM does not support vectors of pointers, so turn any pointers into
     // integers.
-    if (POINTER_TYPE_P(TREE_TYPE(type)))
+    if (isa<ACCESS_TYPE>(TREE_TYPE(type)))
       Ty = getTargetData().getIntPtrType(Context);
     else
-      Ty = ConvertTypeNonRecursive(TYPE_MAIN_VARIANT(TREE_TYPE(type)));
+      Ty = ConvertTypeNonRecursive(main_type(type));
     Ty = VectorType::get(Ty, TYPE_VECTOR_SUBPARTS(type));
     return RememberTypeConversion(type, Ty);
   }
@@ -1528,6 +1536,11 @@ namespace llvm {
 Type *ConvertType(tree type) {
   if (type == error_mark_node) return Type::getInt32Ty(Context);
 
+  // Check that the type mode doesn't depend on the type variant (various bits
+  // of the plugin rely on this).
+  assert(TYPE_MODE(type) == TYPE_MODE(TYPE_MAIN_VARIANT(type))
+         && "Type mode differs between variants!");
+
   // LLVM doesn't care about variants such as const, volatile, or restrict.
   type = TYPE_MAIN_VARIANT(type);
 
@@ -1558,11 +1571,9 @@ Type *ConvertType(tree type) {
     // in the SCC.  This way, if we have both "struct S" and "struct S*" in the
     // SCC then we can return an LLVM "%struct.s*" for the pointer rather than
     // the nasty {}* type we are obliged to return in general.
-    for (unsigned i = 0, e = SCC.size(); i != e; ++i) {
+    for (size_t i = 0, e = SCC.size(); i != e; ++i) {
       tree some_type = SCC[i];
-      if (TREE_CODE(some_type) != QUAL_UNION_TYPE &&
-          TREE_CODE(some_type) != RECORD_TYPE &&
-          TREE_CODE(some_type) != UNION_TYPE) {
+      if (!isa<RECORD_OR_UNION_TYPE>(some_type)) {
         assert(!getCachedType(some_type) && "Type already converted!");
         continue;
       }
@@ -1591,7 +1602,7 @@ Type *ConvertType(tree type) {
     // edges like this destroys all cycles allowing the other types in the SCC
     // to be converted straightforwardly.
     SCCInProgress = &SCC;
-    for (unsigned i = 0, e = SCC.size(); i != e; ++i)
+    for (size_t i = 0, e = SCC.size(); i != e; ++i)
       ConvertType(SCC[i]);
     SCCInProgress = 0;
 
@@ -1601,13 +1612,13 @@ Type *ConvertType(tree type) {
     // uses by types outside the SCC will see a sensible pointer type.  This is
     // not needed for correctness - it just makes the IR nicer.
     if (SCC.size() > 1)
-      for (unsigned i = 0, e = SCC.size(); i != e; ++i) {
+      for (size_t i = 0, e = SCC.size(); i != e; ++i) {
         tree some_type = SCC[i];
-        if (POINTER_TYPE_P(some_type)) {
-          tree pointee = TYPE_MAIN_VARIANT(TREE_TYPE(some_type));
+        if (isa<ACCESS_TYPE>(some_type)) {
+          tree pointee = main_type(some_type);
           // The pointee cannot have void type since the SCC contains more than
           // one type.
-          assert(!VOID_TYPE_P(pointee) && "Self-recursive void*!");
+          assert(!isa<VOID_TYPE>(pointee) && "Self-recursive void*!");
           // The pointee must have been converted since it has to be in the same
           // SCC as the pointer (since the SCC contains more than one type).
           Type *PointeeTy = getCachedType(pointee);
