@@ -1,7 +1,6 @@
 //===--------------- Target.cpp - Implements the IA-32 ABI. ---------------===//
 //
-// Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011  Evan Cheng,
-// Duncan Sands et al.
+// Copyright (C) 2005 to 2012  Evan Cheng, Duncan Sands et al.
 //
 // This file is part of DragonEgg.
 //
@@ -32,7 +31,10 @@
 #include <gmp.h>
 
 // GCC headers
+#include "auto-host.h"
+#ifndef ENABLE_BUILD_WITH_CXX
 extern "C" {
+#endif
 #include "config.h"
 // Stop GCC declaring 'getopt' as it can clash with the system's declaration.
 #undef HAVE_DECL_GETOPT
@@ -43,8 +45,21 @@ extern "C" {
 
 #include "diagnostic.h"
 #include "gimple.h"
+#if (GCC_MINOR > 6)
+#include "gimple-pretty-print.h"
+#endif
 #include "toplev.h"
-}
+
+#if (GCC_MINOR == 6)
+extern void debug_gimple_stmt(union gimple_statement_d *);
+#endif
+
+#ifndef ENABLE_BUILD_WITH_CXX
+} // extern "C"
+#endif
+
+// Trees header.
+#include "dragonegg/Trees.h"
 
 static LLVMContext &Context = getGlobalContext();
 
@@ -61,6 +76,8 @@ static Value *BitCastToIntVector(Value *Op, LLVMBuilder &Builder) {
 /// BuiltinCode - A enumerated type with one value for each supported builtin.
 enum BuiltinCode {
   SearchForHandler, // Builtin not seen before - search for a handler.
+  clzs, // Builtin with exceptional name.
+  ctzs, // Builtin with exceptional name.
 #define DEFINE_BUILTIN(x) x
 #include "x86_builtins"
 #undef DEFINE_BUILTIN
@@ -104,9 +121,11 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
     // No associated BuiltinCode.  Work out what value to use based on the
     // builtin's name.
 
-    // List of builtin names (w/o '__builtin_ia32_') and associated BuiltinCode.
+    // List of builtin names and associated BuiltinCode.
     static const HandlerEntry Handlers[] = {
-#define DEFINE_BUILTIN(x) {#x, x}
+      {"__builtin_clzs", clzs}, // Builtin with exceptional name.
+      {"__builtin_ctzs", ctzs}, // Builtin with exceptional name.
+#define DEFINE_BUILTIN(x) {"__builtin_ia32_" #x, x}
 #include "x86_builtins"
 #undef DEFINE_BUILTIN
     };
@@ -123,14 +142,11 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
 
     Handler = UnsupportedBuiltin;
     const char *Identifier = IDENTIFIER_POINTER(DECL_NAME(fndecl));
-    // All builtins handled here have a name starting with __builtin_ia32_.
-    if (!strncmp(Identifier, "__builtin_ia32_", 15)) {
-      HandlerEntry ToFind = { Identifier + 15, SearchForHandler };
-      const HandlerEntry *E = std::lower_bound(Handlers, Handlers + N, ToFind,
-                                               HandlerLT);
-      if ((E < Handlers + N) && !strcmp(E->Name, ToFind.Name))
-        Handler = E->Handler;
-    }
+    HandlerEntry ToFind = { Identifier, SearchForHandler };
+    const HandlerEntry *E = std::lower_bound(Handlers, Handlers + N, ToFind,
+                                             HandlerLT);
+    if ((E < Handlers + N) && !strcmp(E->Name, ToFind.Name))
+      Handler = E->Handler;
   }
 
   bool flip = false;
@@ -138,7 +154,8 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
 
   switch (Handler) {
   case SearchForHandler:
-    assert(false && "Unexpected builtin code!");
+    debug_gimple_stmt(stmt);
+    llvm_unreachable("Unexpected builtin code!");
   case UnsupportedBuiltin: return false;
   case addps:
   case addps256:
@@ -461,6 +478,8 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
     return true;
   }
   case loadhpd: {
+    PointerType *f64Ptr = Type::getDoublePtrTy(Context);
+    Ops[1] = Builder.CreateBitCast(Ops[1], f64Ptr);
     Value *Load = Builder.CreateLoad(Ops[1]);
     Ops[1] = BuildVector(Load, UndefValue::get(Type::getDoubleTy(Context)), NULL);
     Ops[1] = Builder.CreateBitCast(Ops[1], ResultType);
@@ -469,6 +488,8 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
     return true;
   }
   case loadlpd: {
+    PointerType *f64Ptr = Type::getDoublePtrTy(Context);
+    Ops[1] = Builder.CreateBitCast(Ops[1], f64Ptr);
     Value *Load = Builder.CreateLoad(Ops[1]);
     Ops[1] = BuildVector(Load, UndefValue::get(Type::getDoubleTy(Context)), NULL);
     Ops[1] = Builder.CreateBitCast(Ops[1], ResultType);
@@ -674,6 +695,7 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
 
         Value* SV = ConstantVector::get(Indices);
         Result = Builder.CreateShuffleVector(Ops[1], Ops[0], SV, "palignr");
+        Result = Builder.CreateBitCast(Result, ResultType);
         return true;
       }
 
@@ -683,15 +705,18 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
         // MMX has these as 1 x i64 vectors for some odd optimization reasons.
         Type *EltTy = Type::getInt64Ty(Context);
         Type *VecTy = VectorType::get(EltTy, 1);
+        Type *MMXTy = Type::getX86_MMXTy(Context);
 
-        Ops[0] = Builder.CreateBitCast(Ops[0], VecTy, "cast");
+        Ops[0] = Builder.CreateBitCast(Ops[0], MMXTy);
         Ops[1] = ConstantInt::get(VecTy, (shiftVal-8) * 8);
+        Ops[1] = Builder.CreateBitCast(Ops[1], MMXTy);
 
         // create i32 constant
         Function *F = Intrinsic::getDeclaration(TheModule,
                                                 Intrinsic::x86_mmx_psrl_q);
         Result = Builder.CreateCall(F, ArrayRef<Value *>(&Ops[0], 2),
                                     "palignr");
+        Result = Builder.CreateBitCast(Result, ResultType);
         return true;
       }
 
@@ -727,6 +752,7 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
 
         Value* SV = ConstantVector::get(Indices);
         Result = Builder.CreateShuffleVector(Ops[1], Ops[0], SV, "palignr");
+        Result = Builder.CreateBitCast(Result, ResultType);
         return true;
       }
 
@@ -745,6 +771,7 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
                                                 Intrinsic::x86_sse2_psrl_dq);
         Result = Builder.CreateCall(F, ArrayRef<Value *>(&Ops[0], 2),
                                     "palignr");
+        Result = Builder.CreateBitCast(Result, ResultType);
         return true;
       }
 
@@ -859,9 +886,60 @@ bool TreeToLLVM::TargetIntrinsicLower(gimple stmt,
     Result = Builder.CreateShuffleVector(Ops[0], Ops[1], Mask);
     return true;
   }
+  case pcmpeqb128:
+  case pcmpeqb256:
+  case pcmpeqd128:
+  case pcmpeqd256:
+  case pcmpeqq:
+  case pcmpeqq256:
+  case pcmpeqw128:
+  case pcmpeqw256:
+    Result = Builder.CreateICmpEQ(Ops[0], Ops[1]);
+    // Need to sign extend since icmp returns a vector of i1.
+    Result = Builder.CreateSExt(Result, ResultType);
+    return true;
+  case pcmpgtb128:
+  case pcmpgtb256:
+  case pcmpgtd128:
+  case pcmpgtd256:
+  case pcmpgtq:
+  case pcmpgtq256:
+  case pcmpgtw128:
+  case pcmpgtw256:
+    Result = Builder.CreateICmpSGT(Ops[0], Ops[1]);
+    // Need to sign extend since icmp returns a vector of i1.
+    Result = Builder.CreateSExt(Result, ResultType);
+    return true;
+  case pswapdsf:
+  case pswapdsi: {
+    Type *MMXTy = Type::getX86_MMXTy(Context);
+    Ops[0] = Builder.CreateBitCast(Ops[0], MMXTy);
+    Function *pswapd = Intrinsic::getDeclaration(TheModule,
+                                                 Intrinsic::x86_3dnowa_pswapd);
+    Result = Builder.CreateCall(pswapd, Ops[0]);
+    Result = Builder.CreateBitCast(Result, ResultType);
+    return true;
   }
-  DieAbjectly("Builtin not implemented!", stmt);
-  return false;
+  case clzs: {
+    // The value is usually passed in as an int rather than as a short.
+    Type *Int16Ty = Builder.getInt16Ty();
+    Result = Builder.CreateTruncOrBitCast(Ops[0], Int16Ty);
+    Function *ctlz = Intrinsic::getDeclaration(TheModule, Intrinsic::ctlz,
+                                               Int16Ty);
+    Result = Builder.CreateCall2(ctlz, Result, Builder.getTrue());
+    return true;
+  }
+  case ctzs: {
+    // The value is usually passed in as an int rather than as a short.
+    Type *Int16Ty = Builder.getInt16Ty();
+    Result = Builder.CreateTruncOrBitCast(Ops[0], Int16Ty);
+    Function *cttz = Intrinsic::getDeclaration(TheModule, Intrinsic::cttz,
+                                               Int16Ty);
+    Result = Builder.CreateCall2(cttz, Result, Builder.getTrue());
+    return true;
+  }
+  }
+  llvm_unreachable("Forgot case for code?");
 }
 
 // One day we will do parameter marshalling right: by using CUMULATIVE_ARGS.
@@ -942,7 +1020,7 @@ llvm_x86_32_should_pass_aggregate_in_mixed_regs(tree TreeType, Type *Ty,
 /* It returns true if an aggregate of the specified type should be passed as a
    first class aggregate. */
 bool llvm_x86_should_pass_aggregate_as_fca(tree type, Type *Ty) {
-  if (TREE_CODE(type) != COMPLEX_TYPE)
+  if (!isa<COMPLEX_TYPE>(type))
     return false;
   StructType *STy = dyn_cast<StructType>(Ty);
   if (!STy || STy->isPacked()) return false;
@@ -982,7 +1060,7 @@ bool llvm_x86_should_pass_aggregate_in_memory(tree TreeType, Type *Ty) {
    register used so far.  Caller is responsible for initializing outputs. */
 static void count_num_registers_uses(std::vector<Type*> &ScalarElts,
                                      unsigned &NumGPRs, unsigned &NumXMMs) {
-  for (unsigned i = 0, e = ScalarElts.size(); i != e; ++i) {
+  for (size_t i = 0, e = ScalarElts.size(); i != e; ++i) {
     Type *Ty = ScalarElts[i];
     if (VectorType *VTy = dyn_cast<VectorType>(Ty)) {
       if (!TARGET_MACHO)
@@ -1104,15 +1182,15 @@ llvm_x86_64_should_pass_aggregate_in_mixed_regs(tree TreeType, Type *Ty,
           Elts.push_back (Type::getFloatTy(Context));
           Bytes -= 4;
         } else
-          assert(0 && "Not yet handled!");
+          llvm_unreachable("Not yet handled!");
       } else if ((NumClasses-i) == 2) {
         if (Class[i+1] == X86_64_SSEUP_CLASS) {
-          Type *Ty = ConvertType(TreeType);
-          if (StructType *STy = dyn_cast<StructType>(Ty))
+          Type *LLVMTy = ConvertType(TreeType);
+          if (StructType *STy = dyn_cast<StructType>(LLVMTy))
             // Look pass the struct wrapper.
             if (STy->getNumElements() == 1)
-              Ty = STy->getElementType(0);
-          if (VectorType *VTy = dyn_cast<VectorType>(Ty)) {
+              LLVMTy = STy->getElementType(0);
+          if (VectorType *VTy = dyn_cast<VectorType>(LLVMTy)) {
             if (VTy->getNumElements() == 2) {
               if (VTy->getElementType()->isIntegerTy()) {
                 Elts.push_back(VectorType::get(Type::getInt64Ty(Context), 2));
@@ -1129,7 +1207,7 @@ llvm_x86_64_should_pass_aggregate_in_mixed_regs(tree TreeType, Type *Ty,
               }
               Bytes -= 4;
             }
-          } else if (llvm_x86_is_all_integer_types(Ty)) {
+          } else if (llvm_x86_is_all_integer_types(LLVMTy)) {
             Elts.push_back(VectorType::get(Type::getInt32Ty(Context), 4));
             Bytes -= 4;
           } else {
@@ -1157,10 +1235,10 @@ llvm_x86_64_should_pass_aggregate_in_mixed_regs(tree TreeType, Type *Ty,
           Elts.push_back(Type::getVoidTy(Context));
           Bytes -= 16;
         } else
-          assert(0 && "Not yet handled!");
+          llvm_unreachable("Not yet handled!");
         ++i; // Already handled the next one.
       } else
-        assert(0 && "Not yet handled!");
+        llvm_unreachable("Not yet handled!");
       break;
     case X86_64_SSESF_CLASS:
       totallyEmpty = false;
@@ -1182,7 +1260,7 @@ llvm_x86_64_should_pass_aggregate_in_mixed_regs(tree TreeType, Type *Ty,
       Elts.push_back(Type::getVoidTy(Context));
       Bytes -= 8;
       break;
-    default: assert(0 && "Unexpected register class!");
+    default: llvm_unreachable("Unexpected register class!");
     }
   }
 
@@ -1195,9 +1273,9 @@ llvm_x86_64_should_pass_aggregate_in_mixed_regs(tree TreeType, Type *Ty,
 bool llvm_x86_should_pass_vector_in_integer_regs(tree type) {
   if (!TARGET_MACHO)
     return false;
-  if (TREE_CODE(type) == VECTOR_TYPE &&
+  if (isa<VECTOR_TYPE>(type) &&
       TYPE_SIZE(type) &&
-      TREE_CODE(TYPE_SIZE(type))==INTEGER_CST) {
+      isa<INTEGER_CST>(TYPE_SIZE(type))) {
     if (TREE_INT_CST_LOW(TYPE_SIZE(type))==64 && TARGET_MMX)
       return false;
     if (TREE_INT_CST_LOW(TYPE_SIZE(type))==128 && TARGET_SSE)
@@ -1215,9 +1293,9 @@ bool llvm_x86_should_pass_vector_using_byval_attr(tree type) {
     return false;
   if (!TARGET_64BIT)
     return false;
-  if (TREE_CODE(type) == VECTOR_TYPE &&
+  if (isa<VECTOR_TYPE>(type) &&
       TYPE_SIZE(type) &&
-      TREE_CODE(TYPE_SIZE(type))==INTEGER_CST) {
+      isa<INTEGER_CST>(TYPE_SIZE(type))) {
     if (TREE_INT_CST_LOW(TYPE_SIZE(type))<=128)
       return false;
   }
@@ -1233,9 +1311,9 @@ bool llvm_x86_should_pass_vector_using_byval_attr(tree type) {
 tree llvm_x86_should_return_vector_as_scalar(tree type, bool isBuiltin) {
   if (TARGET_MACHO &&
       !isBuiltin &&
-      TREE_CODE(type) == VECTOR_TYPE &&
+      isa<VECTOR_TYPE>(type) &&
       TYPE_SIZE(type) &&
-      TREE_CODE(TYPE_SIZE(type))==INTEGER_CST) {
+      isa<INTEGER_CST>(TYPE_SIZE(type))) {
     if (TREE_INT_CST_LOW(TYPE_SIZE(type))==64 &&
         TYPE_VECTOR_SUBPARTS(type)==1)
       return uint64_type_node;
@@ -1255,9 +1333,9 @@ tree llvm_x86_should_return_selt_struct_as_scalar(tree type) {
   tree retType = isSingleElementStructOrArray(type, true, false);
   if (!retType || !TARGET_64BIT || !TARGET_MACHO)
     return retType;
-  if (TREE_CODE(retType) == VECTOR_TYPE &&
+  if (isa<VECTOR_TYPE>(retType) &&
       TYPE_SIZE(retType) &&
-      TREE_CODE(TYPE_SIZE(retType))==INTEGER_CST &&
+      isa<INTEGER_CST>(TYPE_SIZE(retType)) &&
       TREE_INT_CST_LOW(TYPE_SIZE(retType))==64)
     return double_type_node;
   return retType;
@@ -1269,9 +1347,9 @@ bool llvm_x86_should_return_vector_as_shadow(tree type, bool isBuiltin) {
   if (TARGET_MACHO &&
     !isBuiltin &&
     !TARGET_64BIT &&
-    TREE_CODE(type) == VECTOR_TYPE &&
+    isa<VECTOR_TYPE>(type) &&
     TYPE_SIZE(type) &&
-    TREE_CODE(TYPE_SIZE(type))==INTEGER_CST) {
+    isa<INTEGER_CST>(TYPE_SIZE(type))) {
     if (TREE_INT_CST_LOW(TYPE_SIZE(type))==64 &&
        TYPE_VECTOR_SUBPARTS(type)>1)
       return true;
@@ -1288,7 +1366,7 @@ bool llvm_x86_should_not_return_complex_in_memory(tree type) {
   if (!TARGET_64BIT)
     return false;
 
-  if (TREE_CODE(type) == COMPLEX_TYPE &&
+  if (isa<COMPLEX_TYPE>(type) &&
       TREE_INT_CST_LOW(TYPE_SIZE_UNIT(type)) == 32)
     return true;
 
@@ -1337,7 +1415,7 @@ static bool llvm_suitable_multiple_ret_value_type(Type *Ty,
 Type *llvm_x86_scalar_type_for_struct_return(tree type, unsigned *Offset) {
   *Offset = 0;
   Type *Ty = ConvertType(type);
-  unsigned Size = getTargetData().getTypeAllocSize(Ty);
+  uint64_t Size = getTargetData().getTypeAllocSize(Ty);
   if (Size == 0)
     return Type::getVoidTy(Context);
   else if (Size == 1)
@@ -1376,7 +1454,7 @@ Type *llvm_x86_scalar_type_for_struct_return(tree type, unsigned *Offset) {
         else
           return Type::getInt8Ty(Context);
       }
-      assert(0 && "Unexpected type!");
+      llvm_unreachable("Unexpected type!");
     }
     if (NumClasses == 2) {
       if (Class[1] == X86_64_NO_CLASS) {
@@ -1388,7 +1466,7 @@ Type *llvm_x86_scalar_type_for_struct_return(tree type, unsigned *Offset) {
           return Type::getDoubleTy(Context);
         else if (Class[0] == X86_64_SSESF_CLASS)
           return Type::getFloatTy(Context);
-        assert(0 && "Unexpected type!");
+        llvm_unreachable("Unexpected type!");
       }
       if (Class[0] == X86_64_NO_CLASS) {
         *Offset = 8;
@@ -1399,11 +1477,11 @@ Type *llvm_x86_scalar_type_for_struct_return(tree type, unsigned *Offset) {
           return Type::getDoubleTy(Context);
         else if (Class[1] == X86_64_SSESF_CLASS)
           return Type::getFloatTy(Context);
-        assert(0 && "Unexpected type!");
+        llvm_unreachable("Unexpected type!");
       }
-      assert(0 && "Unexpected type!");
+      llvm_unreachable("Unexpected type!");
     }
-    assert(0 && "Unexpected type!");
+    llvm_unreachable("Unexpected type!");
   } else {
     if (Size <= 8)
       return Type::getInt64Ty(Context);
@@ -1420,7 +1498,7 @@ Type *llvm_x86_scalar_type_for_struct_return(tree type, unsigned *Offset) {
 /// This routine uses GCC implementation to find required register classes.
 /// The original implementation of this routine is based on
 /// llvm_x86_64_should_pass_aggregate_in_mixed_regs code.
-void
+static void
 llvm_x86_64_get_multiple_return_reg_classes(tree TreeType, Type * /*Ty*/,
                                             std::vector<Type*> &Elts) {
   enum x86_64_reg_class Class[MAX_CLASSES];
@@ -1428,15 +1506,13 @@ llvm_x86_64_get_multiple_return_reg_classes(tree TreeType, Type * /*Ty*/,
   HOST_WIDE_INT Bytes =
     (Mode == BLKmode) ? int_size_in_bytes(TreeType) : (int) GET_MODE_SIZE(Mode);
   int NumClasses = classify_argument(Mode, TreeType, Class, 0);
-  if (!NumClasses)
-     assert(0 && "This type does not need multiple return registers!");
+  assert(NumClasses && "This type does not need multiple return registers!");
 
-  if (NumClasses == 1 && Class[0] == X86_64_INTEGERSI_CLASS)
-    // This will fit in one i32 register.
-     assert(0 && "This type does not need multiple return registers!");
+  assert((NumClasses != 1 || Class[0] != X86_64_INTEGERSI_CLASS) &&
+         "This will fit in one i32 register!");
 
-  if (NumClasses == 1 && Class[0] == X86_64_INTEGER_CLASS)
-     assert(0 && "This type does not need multiple return registers!");
+  assert((NumClasses != 1 || Class[0] != X86_64_INTEGER_CLASS) &&
+         "This type does not need multiple return registers!");
 
   // classify_argument uses a single X86_64_NO_CLASS as a special case for
   // empty structs. Recognize it and don't add any return values in that
@@ -1468,7 +1544,7 @@ llvm_x86_64_get_multiple_return_reg_classes(tree TreeType, Type * /*Ty*/,
           Elts.push_back(Type::getFloatTy(Context));
           Bytes -= 4;
         } else
-          assert(0 && "Not yet handled!");
+          llvm_unreachable("Not yet handled!");
       } else if ((NumClasses-i) == 2) {
         if (Class[i+1] == X86_64_SSEUP_CLASS) {
           Type *Ty = ConvertType(TreeType);
@@ -1517,11 +1593,11 @@ llvm_x86_64_get_multiple_return_reg_classes(tree TreeType, Type * /*Ty*/,
           Elts.push_back(Type::getDoubleTy(Context));
           Bytes -= 16;
         } else {
-          assert(0 && "Not yet handled!");
+          llvm_unreachable("Not yet handled!");
         }
         ++i; // Already handled the next one.
       } else
-        assert(0 && "Not yet handled!");
+        llvm_unreachable("Not yet handled!");
       break;
     case X86_64_SSESF_CLASS:
       Elts.push_back(Type::getFloatTy(Context));
@@ -1540,7 +1616,7 @@ llvm_x86_64_get_multiple_return_reg_classes(tree TreeType, Type * /*Ty*/,
       // padding bytes.
       Elts.push_back(Type::getInt64Ty(Context));
       break;
-    default: assert(0 && "Unexpected register class!");
+    default: llvm_unreachable("Unexpected register class!");
     }
   }
 }
